----------------------------------
-- Variables
----------------------------------
local TCS =
{
    -- AddOn's Controller
    control = nil,
    labelControl = nil,

    -- Constants
    NAME            = "TsCombatStats",
    SV              = "TCS_SV_r4",

    TOOLTIP_FONT    = "ZoFontWinH4",

    -- AddOn Settings
    settingsMenu    = nil,
    settings        =
    {
        chatTabOptions      = {},
        abilityStatistics   = {},

        abilityStatisticOptions =
        {
            enableTooltip       = true,

            showNormalAverage   = true,
            showNormalMaximum   = true,
            showNormalMinimum   = true,

            showCriticalAverage = true,
            showCriticalMaximum = true,
            showCriticalMinimum = true,

            showTotalAverage    = true,
            showCriticalPercent = true
        },

        alliancePointsMinimum                   = 5,
        enableCombatSessionDisplay              = true,
        showDPS                                 = true,
        totalDamage                             = true,
        showHPS                                 = true,
        totalHealing                            = true,
        combatSessionDisplayBGType              = "",
        combatSessionDisplayScale               = 1,
        combatSessionDisplayBGTransparency      = 100,
        combatSessionDisplayTextTransparency    = 100,

        combatSessionDPSDisplayPoint            = BOTTOM,
        combatSessionDPSDisplayRelativeTo       = "GuiRoot",
        combatSessionDPSDisplayRelativePoint    = BOTTOM,
        combatSessionDPSDisplayOffsetX          = -70,
        combatSessionDPSDisplayOffsetY          = -73,

        combatSessionTotalDamageDisplayPoint            = BOTTOM,
        combatSessionTotalDamageDisplayRelativeTo       = "GuiRoot",
        combatSessionTotalDamageDisplayRelativePoint    = BOTTOM,
        combatSessionTotalDamageDisplayOffsetX          = -210,
        combatSessionTotalDamageDisplayOffsetY          = -73,

        combatSessionHPSDisplayPoint            = BOTTOM,
        combatSessionHPSDisplayRelativeTo       = "GuiRoot",
        combatSessionHPSDisplayRelativePoint    = BOTTOM,
        combatSessionHPSDisplayOffsetX          = 70,
        combatSessionHPSDisplayOffsetY          = -73,

        combatSessionTotalHealingDisplayPoint           = BOTTOM,
        combatSessionTotalHealingDisplayRelativeTo      = "GuiRoot",
        combatSessionTotalHealingDisplayRelativePoint   = BOTTOM,
        combatSessionTotalHealingDisplayOffsetX         = 210,
        combatSessionTotalHealingDisplayOffsetY         = -73,

        combatSessionAbilityDamageEnableBarIndicators   = true,
        combatSessionAbilityDamageBarWidth              = 15,
        combatSessionAbilityDamageBarR                  = 0.8039,
        combatSessionAbilityDamageBarG                  = 0.0862,
        combatSessionAbilityDamageBarB                  = 0.0862,
        combatSessionAbilityDamageBarAlignment          = "",
        combatSessionAbilityDamageEnablePercentage      = true,
        combatSessionAbilityDamagePercentageSize        = 1,
        combatSessionAbilityDamagePercentageAlignment   = "",
    },

    damageTypeStrings =
    {
        [DAMAGE_TYPE_COLD]        = "SI_TCS_DAMAGE_TYPE_COLD",
        [DAMAGE_TYPE_DISEASE]     = "SI_TCS_DAMAGE_TYPE_DISEASE",
        [DAMAGE_TYPE_DROWN]       = "SI_TCS_DAMAGE_TYPE_DROWN",
        [DAMAGE_TYPE_EARTH]       = "SI_TCS_DAMAGE_TYPE_EARTH",
        [DAMAGE_TYPE_FIRE]        = "SI_TCS_DAMAGE_TYPE_FIRE",
        [DAMAGE_TYPE_GENERIC]     = "SI_TCS_DAMAGE_TYPE_GENERIC",
        [DAMAGE_TYPE_MAGIC]       = "SI_TCS_DAMAGE_TYPE_MAGIC",
        [DAMAGE_TYPE_NONE]        = "SI_TCS_DAMAGE_TYPE_NONE",
        [DAMAGE_TYPE_OBLIVION]    = "SI_TCS_DAMAGE_TYPE_OBLIVION",
        [DAMAGE_TYPE_PHYSICAL]    = "SI_TCS_DAMAGE_TYPE_PHYSICAL",
        [DAMAGE_TYPE_POISON]      = "SI_TCS_DAMAGE_TYPE_POISON",
        [DAMAGE_TYPE_SHOCK]       = "SI_TCS_DAMAGE_TYPE_SHOCK",
    },

    -- Last MouseOver
    lastMouseOver               = nil,

    -- Skill/Ability Cache
    abilityInfoCache = {},

    -- Combat session
    isInCombat = false,

    combatSessionStartTime = 0,

    combatSessionSkillAbilityExperienceGained = {},
    combatSessionAbilityDamage = {},

    combatSessionDPSDisplay             = nil,
    combatSessionTotalDamageDisplay     = nil,
    combatSessionHPSDisplay             = nil,
    combatSessionTotalHealingDisplay    = nil,

    combatSessionExperienceGained                   = 0,
    combatSessionFinesseBonusExperiencePercentage   = 0,

    combatSessionDamageDealt    = 0,
    combatSessionHealingDone    = 0,

    -- Misc vars
    chatTabFilterCount          = 0,
    chatTabFilterRowCount       = 0,
    chatTabFilterColumnCount    = 0,

    toolTipInitializeCounter    = 0,
    lastExperiences             = {},
    onlyEnable                  = false,

    abilityNameProxies = {},
    slotNameCache = {},
}

----------------------------------
-- Core
----------------------------------
function TCS_Initialize( _control )
    TCS.control = _control
    TCS:OnInitialized()
end

function TCS:OnInitialized()
    -- Register events
    EVENT_MANAGER:RegisterForEvent( self.NAME, EVENT_ADD_ON_LOADED, function( ... ) self:OnEventAddOnLoaded( ... ) end )
end

function TCS:OnEventAddOnLoaded( _event, _addonName )
    -- If the addon is not ours, do nothing
    if( _addonName ~= self.NAME ) then
        return
    end

    -- Create a dummy label for the use of GetStringWidth()
    self.labelControl = WINDOW_MANAGER:CreateControl( self.NAME .. "DummyLabel", GuiRoot, CT_LABEL )
    self.labelControl:SetFont( self.TOOLTIP_FONT )
    self.labelControl:SetHidden( true )

    -- Initialize saved vars
    self.settings.combatSessionAbilityDamageBarAlignment = GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_BAR_ALIGNMENT_LEFT )
    self.settings.combatSessionAbilityDamagePercentageAlignment = GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_BOTTOMLEFT )
    self.settings.combatSessionDisplayBGType = GetString( SI_TCS_SETTINGS_BACKGROUND_TYPE_A )
    self.settings = ZO_SavedVars:New( self.SV, 2, nil, self.settings )

    -- Initialize GUI and Settings
    self:InitializeAddOnMenu()
    self:ChatTabOptionsIntialize()
    self:CombatSessionDisplayInitialize()
    self:ChatCombatOutputInitialize()
    self:SkillTooltipInitialize()
    self:InitializeAbilityDamagePercentages()

    -- Register handlers
    ZO_PreHookHandler(
        self.control, "OnUpdate",
        function() self:OnUpdate() end
    )

    ZO_PreHookHandler(
        ZO_Skills, "OnHide",
        function() self:HideCombatSessionDisplays( false ) end
    )

    ZO_PreHookHandler(
        ZO_Skills, "OnShow",
        function() self:HideCombatSessionDisplays( true ) end
    )

    ZO_PreHookHandler(
        TCS_Combat_Session_DPS_Display, "OnMoveStop",
        function()
            local isValidAnchor, point, relativeTo, relativePoint, offsetX, offsetY = TCS_Combat_Session_DPS_Display:GetAnchor(0)

            self.settings.combatSessionDPSDisplayPoint          = point
            self.settings.combatSessionDPSDisplayRelativeTo     = relativeTo:GetName()
            self.settings.combatSessionDPSDisplayRelativePoint  = relativePoint
            self.settings.combatSessionDPSDisplayOffsetX        = offsetX
            self.settings.combatSessionDPSDisplayOffsetY        = offsetY
        end
    )

    ZO_PreHookHandler(
        TCS_Combat_Session_Total_Damage_Display, "OnMoveStop",
        function()
            local isValidAnchor, point, relativeTo, relativePoint, offsetX, offsetY = TCS_Combat_Session_Total_Damage_Display:GetAnchor(0)

            self.settings.combatSessionTotalDamageDisplayPoint          = point
            self.settings.combatSessionTotalDamageDisplayRelativeTo     = relativeTo:GetName()
            self.settings.combatSessionTotalDamageDisplayRelativePoint  = relativePoint
            self.settings.combatSessionTotalDamageDisplayOffsetX        = offsetX
            self.settings.combatSessionTotalDamageDisplayOffsetY        = offsetY
        end
    )

    ZO_PreHookHandler(
        TCS_Combat_Session_HPS_Display, "OnMoveStop",
        function()
            local isValidAnchor, point, relativeTo, relativePoint, offsetX, offsetY = TCS_Combat_Session_HPS_Display:GetAnchor(0)

            self.settings.combatSessionHPSDisplayPoint          = point
            self.settings.combatSessionHPSDisplayRelativeTo     = relativeTo:GetName()
            self.settings.combatSessionHPSDisplayRelativePoint  = relativePoint
            self.settings.combatSessionHPSDisplayOffsetX        = offsetX
            self.settings.combatSessionHPSDisplayOffsetY        = offsetY
        end
    )

    ZO_PreHookHandler(
        TCS_Combat_Session_Total_Healing_Display, "OnMoveStop",
        function()
            local isValidAnchor, point, relativeTo, relativePoint, offsetX, offsetY = TCS_Combat_Session_Total_Healing_Display:GetAnchor(0)

            self.settings.combatSessionTotalHealingDisplayPoint         = point
            self.settings.combatSessionTotalHealingDisplayRelativeTo    = relativeTo:GetName()
            self.settings.combatSessionTotalHealingDisplayRelativePoint = relativePoint
            self.settings.combatSessionTotalHealingDisplayOffsetX       = offsetX
            self.settings.combatSessionTotalHealingDisplayOffsetY       = offsetY
        end
    )

    -- Register chat commands
    SLASH_COMMANDS[ "/clear" ] =
        function( _param )
            if( _param == "all" ) then
                for i = 1, GetNumChatContainers() do
                    for j = 1, GetNumChatContainerTabs( i ) do
                        CHAT_SYSTEM.containers[ i ].windows[ j ].buffer:Clear()
                    end
                end
            else
                CHAT_SYSTEM.primaryContainer.currentBuffer:Clear()
            end
        end

end

function TCS:InitializeAddOnMenu()
    local LAM = LibStub('LibAddonMenu-2.0')
    local LAMCC = LAMCreateControl
    local panel, lastControl

    local function AddControl(widgetData)
        local widgetType = widgetData.type
        local widget = LAMCC[widgetType](panel, widgetData)
        if lastControl then
            widget:SetAnchor(TOPLEFT, lastControl, BOTTOMLEFT, 0, 15)
            lastControl = widget
        else
            widget:SetAnchor(TOPLEFT)
            lastControl = widget
        end
    end

    -- Create the menu
    local panelData = {
        type = "panel",
        name = "T's Combat Stats",
        displayName = ZO_HIGHLIGHT_TEXT:Colorize("T's Combat Stats"),
        author = "Tingle0x539",
        version = "0.17",
        slashCommand = "/tcs",
    }
    panel = LAM:RegisterAddonPanel('TCS_SettingsMenu', panelData)

    -- Header for the ability statistic
    AddControl({ type = "header", name = GetString( SI_TCS_SETTINGS_TOOLTIP_LABEL ) })

    AddControl({
        type = "checkbox",
        name = GetString( SI_TCS_SETTINGS_TOOLTIP_ENABLE_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_TOOLTIP_ENABLE_TOOLTIP ),
        getFunc = function() return self.settings.abilityStatisticOptions.enableTooltip end,
        setFunc = function( _set )
                self.settings.abilityStatisticOptions.enableTooltip         = _set

                local currentMouseOver = WINDOW_MANAGER:GetMouseOverControl()
                if( currentMouseOver == TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox or
                    currentMouseOver == TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox.checkbox
                ) then
                    if( self.settings.abilityStatisticOptions.showNormalAverage ~= _set )   then zo_callHandler( TCS_SettingsMenu_NormalAverage_Checkbox,   "OnMouseUp" ) end
                    if( self.settings.abilityStatisticOptions.showNormalMaximum ~= _set )   then zo_callHandler( TCS_SettingsMenu_NormalMaximum_Checkbox,   "OnMouseUp" ) end
                    if( self.settings.abilityStatisticOptions.showNormalMinimum ~= _set )   then zo_callHandler( TCS_SettingsMenu_NormalMinimum_Checkbox,   "OnMouseUp" ) end
                    if( self.settings.abilityStatisticOptions.showCriticalAverage ~= _set ) then zo_callHandler( TCS_SettingsMenu_CriticalAverage_Checkbox, "OnMouseUp" ) end
                    if( self.settings.abilityStatisticOptions.showCriticalMaximum ~= _set ) then zo_callHandler( TCS_SettingsMenu_CriticalMaximum_Checkbox, "OnMouseUp" ) end
                    if( self.settings.abilityStatisticOptions.showCriticalMinimum ~= _set ) then zo_callHandler( TCS_SettingsMenu_CriticalMinimum_Checkbox, "OnMouseUp" ) end
                    if( self.settings.abilityStatisticOptions.showTotalAverage ~= _set )    then zo_callHandler( TCS_SettingsMenu_TotalAverage_Checkbox,    "OnMouseUp" ) end
                    if( self.settings.abilityStatisticOptions.showCriticalPercent ~= _set ) then zo_callHandler( TCS_SettingsMenu_CriticalPercent_Checkbox, "OnMouseUp" ) end
                end
            end,
        reference = "TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox",
    })

    -- Normal damage
    AddControl({
        type = "checkbox",
        name = "     " .. GetString( SI_TCS_SETTINGS_AVERAGE_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_AVERAGE_TOOLTIP ),
        getFunc = function() return self.settings.abilityStatisticOptions.showNormalAverage end,
        setFunc = function( _set )
                self.settings.abilityStatisticOptions.showNormalAverage = _set

                local currentMouseOver = WINDOW_MANAGER:GetMouseOverControl()
                if( not self.settings.abilityStatisticOptions.enableTooltip and
                    currentMouseOver ~= TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox and
                    currentMouseOver ~= TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox.checkbox
                ) then
                    zo_callHandler( TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox, "OnMouseUp" )
                end
            end,
        reference = "TCS_SettingsMenu_NormalAverage_Checkbox",
    })

    AddControl({
        type = "checkbox",
        name = "          " .. GetString( SI_TCS_SETTINGS_MAXIMUM_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_MAXIMUM_TOOLTIP ),
        getFunc = function() return self.settings.abilityStatisticOptions.showNormalMaximum end,
        setFunc = function( _set )
                self.settings.abilityStatisticOptions.showNormalMaximum = _set

                local currentMouseOver = WINDOW_MANAGER:GetMouseOverControl()
                if( not self.settings.abilityStatisticOptions.enableTooltip and
                    currentMouseOver ~= TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox and
                    currentMouseOver ~= TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox.checkbox
                ) then
                    zo_callHandler( TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox, "OnMouseUp" )
                end
            end,
        reference = "TCS_SettingsMenu_NormalMaximum_Checkbox",
    })

    AddControl({
        type = "checkbox",
        name = "          " .. GetString( SI_TCS_SETTINGS_MINIMUM_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_MINIMUM_TOOLTIP ),
        getFunc = function() return self.settings.abilityStatisticOptions.showNormalMinimum end,
        setFunc = function( _set )
                self.settings.abilityStatisticOptions.showNormalMinimum = _set

                local currentMouseOver = WINDOW_MANAGER:GetMouseOverControl()
                if( not self.settings.abilityStatisticOptions.enableTooltip and
                    currentMouseOver ~= TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox and
                    currentMouseOver ~= TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox.checkbox
                ) then
                    zo_callHandler( TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox, "OnMouseUp" )
                end
            end,
        reference = "TCS_SettingsMenu_NormalMinimum_Checkbox",
    })

    -- Critical Damage
    AddControl({
        type = "checkbox",
        name = "     " .. GetString( SI_TCS_SETTINGS_CRITICAL_AVERAGE_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_CRITICAL_AVERAGE_TOOLTIP ),
        getFunc = function() return self.settings.abilityStatisticOptions.showCriticalAverage end,
        setFunc = function( _set )
                self.settings.abilityStatisticOptions.showCriticalAverage = _set

                local currentMouseOver = WINDOW_MANAGER:GetMouseOverControl()
                if( not self.settings.abilityStatisticOptions.enableTooltip and
                    currentMouseOver ~= TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox and
                    currentMouseOver ~= TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox.checkbox
                ) then
                    zo_callHandler( TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox, "OnMouseUp" )
                end
            end,
        reference = "TCS_SettingsMenu_CriticalAverage_Checkbox",
    })

    AddControl({
        type = "checkbox",
        name = "          " .. GetString( SI_TCS_SETTINGS_CRITICAL_MAXIMUM_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_CRITICAL_MAXIMUM_TOOLTIP ),
        getFunc = function() return self.settings.abilityStatisticOptions.showCriticalMaximum end,
        setFunc = function( _set )
                self.settings.abilityStatisticOptions.showCriticalMaximum = _set

                local currentMouseOver = WINDOW_MANAGER:GetMouseOverControl()
                if( not self.settings.abilityStatisticOptions.enableTooltip and
                    currentMouseOver ~= TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox and
                    currentMouseOver ~= TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox.checkbox
                ) then
                    zo_callHandler( TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox, "OnMouseUp" )
                end
            end,
        reference = "TCS_SettingsMenu_CriticalMaximum_Checkbox",
    })

    AddControl({
        type = "checkbox",
        name = "          " .. GetString( SI_TCS_SETTINGS_CRITICAL_MINIMUM_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_CRITICAL_MINIMUM_TOOLTIP ),
        getFunc = function() return self.settings.abilityStatisticOptions.showCriticalMinimum end,
        setFunc = function( _set )
                self.settings.abilityStatisticOptions.showCriticalMinimum = _set

                local currentMouseOver = WINDOW_MANAGER:GetMouseOverControl()
                if( not self.settings.abilityStatisticOptions.enableTooltip and
                    currentMouseOver ~= TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox and
                    currentMouseOver ~= TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox.checkbox
                ) then
                    zo_callHandler( TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox, "OnMouseUp" )
                end
            end,
        reference = "TCS_SettingsMenu_CriticalMinimum_Checkbox",
    })

    -- Total Average
    AddControl({
        type = "checkbox",
        name = "     " .. GetString( SI_TCS_SETTINGS_TOTAL_AVERAGE_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_TOTAL_AVERAGE_TOOLTIP ),
        getFunc = function() return self.settings.abilityStatisticOptions.showTotalAverage end,
        setFunc = function( _set )
                self.settings.abilityStatisticOptions.showTotalAverage = _set

                local currentMouseOver = WINDOW_MANAGER:GetMouseOverControl()
                if( not self.settings.abilityStatisticOptions.enableTooltip and
                    currentMouseOver ~= TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox and
                    currentMouseOver ~= TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox.checkbox
                ) then
                    zo_callHandler( TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox, "OnMouseUp" )
                end
            end,
        reference = "TCS_SettingsMenu_TotalAverage_Checkbox",
    })

    -- Critical Percent
    AddControl({
        type = "checkbox",
        name = "     " .. GetString( SI_TCS_SETTINGS_CRITICAL_PERCENT_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_CRITICAL_PERCENT_TOOLTIP ),
        getFunc = function() return self.settings.abilityStatisticOptions.showCriticalPercent end,
        setFunc = function( _set )
                self.settings.abilityStatisticOptions.showCriticalPercent = _set

                local currentMouseOver = WINDOW_MANAGER:GetMouseOverControl()
                if( not self.settings.abilityStatisticOptions.enableTooltip and
                    currentMouseOver ~= TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox and
                    currentMouseOver ~= TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox.checkbox
                ) then
                    zo_callHandler( TCS_SettingsMenu_AbilityStatistic_Enable_Checkbox, "OnMouseUp" )
                end
            end,
        reference = "TCS_SettingsMenu_CriticalPercent_Checkbox",
    })

    -- Header for the combat logging
    AddControl({ type = "description", text = "" })
    AddControl({ type = "header", name = GetString( SI_TCS_SETTINGS_COMBAT_LOGGING_LABEL ) })

    -- Alliance Points
    AddControl({
        type = "slider",
        name = GetString( SI_TCS_SETTINGS_ALLIANCE_POINTS_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_ALLIANCE_POINTS_TOOLTIP ),
        min = 0,
        max = 100,
        getFunc = function() return self.settings.alliancePointsMinimum end,
        setFunc = function( _set ) self.settings.alliancePointsMinimum = _set end,
    })

    -- Header for the combat statistics display
    AddControl({ type = "description", text = "" })
    AddControl({ type = "header", name = GetString( SI_TCS_SETTINGS_COMBAT_SESSION_DISPLAY_GUI ) })

    -- Enable dps/hps element
    AddControl({
        type = "checkbox",
        name = GetString( SI_TCS_SETTINGS_COMBAT_SESSION_DISPLAY_ENABLE_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_COMBAT_SESSION_DISPLAY_ENABLE_TOOLTIP ),
        getFunc = function() return self.settings.enableCombatSessionDisplay end,
        setFunc = function( _set )
                self.settings.enableCombatSessionDisplay = _set

                local currentMouseOver = WINDOW_MANAGER:GetMouseOverControl()
                if( currentMouseOver == TCS_SettingsMenu_Enable_Combat_Session_Display_Checkbox or
                    currentMouseOver == TCS_SettingsMenu_Enable_Combat_Session_Display_Checkbox.checkbox
                ) then
                    if( self.settings.showDPS ~= _set )         then zo_callHandler( TCS_SettingsMenu_Show_DPS_Checkbox,            "OnMouseUp" ) end
                    if( self.settings.totalDamage ~= _set )     then zo_callHandler( TCS_SettingsMenu_Show_Total_Damage_Checkbox,   "OnMouseUp" ) end
                    if( self.settings.showHPS ~= _set )         then zo_callHandler( TCS_SettingsMenu_Show_HPS_Checkbox,            "OnMouseUp" ) end
                    if( self.settings.totalHealing ~= _set )    then zo_callHandler( TCS_SettingsMenu_Total_Healing_Checkbox,       "OnMouseUp" ) end
                end
            end,
        reference = "TCS_SettingsMenu_Enable_Combat_Session_Display_Checkbox",
    })

    -- Show dps
    AddControl({
        type = "checkbox",
        name = "     " .. GetString( SI_TCS_SETTINGS_DPS_SHOW_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_DPS_SHOW_TOOLTIP ),
        getFunc = function() return self.settings.showDPS end,
        setFunc = function( _set )
                self.settings.showDPS = _set

                local currentMouseOver = WINDOW_MANAGER:GetMouseOverControl()
                if( not self.settings.enableCombatSessionDisplay and
                    currentMouseOver ~= TCS_SettingsMenu_Enable_Combat_Session_Display_Checkbox and
                    currentMouseOver ~= TCS_SettingsMenu_Enable_Combat_Session_Display_Checkbox.checkbox
                ) then
                    zo_callHandler( TCS_SettingsMenu_Enable_Combat_Session_Display_Checkbox, "OnMouseUp" )
                end

                self:SetCombatSessionDPSUpdateHidden()
            end,
        reference = "TCS_SettingsMenu_Show_DPS_Checkbox",
    })

    -- show total damage
    AddControl({
        type = "checkbox",
        name = "     " .. GetString( SI_TCS_SETTINGS_DAMAGE_TOTAL_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_DAMAGE_TOTAL_TOOLTIP ),
        getFunc = function() return self.settings.totalDamage end,
        setFunc = function( _set )
                self.settings.totalDamage = _set

                local currentMouseOver = WINDOW_MANAGER:GetMouseOverControl()
                if( not self.settings.enableCombatSessionDisplay and
                    currentMouseOver ~= TCS_SettingsMenu_Enable_Combat_Session_Display_Checkbox and
                    currentMouseOver ~= TCS_SettingsMenu_Enable_Combat_Session_Display_Checkbox.checkbox
                ) then
                    zo_callHandler( TCS_SettingsMenu_Enable_Combat_Session_Display_Checkbox, "OnMouseUp" )
                end

                self:SetCombatSessionTotalDamageUpdateHidden()
            end,
        reference = "TCS_SettingsMenu_Show_Total_Damage_Checkbox",
    })

    -- Show hps
    AddControl({
        type = "checkbox",
        name = "     " .. GetString( SI_TCS_SETTINGS_HPS_SHOW_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_HPS_SHOW_TOOLTIP ),
        getFunc = function() return self.settings.showHPS end,
        setFunc = function( _set )
                self.settings.showHPS = _set

                local currentMouseOver = WINDOW_MANAGER:GetMouseOverControl()
                if( not self.settings.enableCombatSessionDisplay and
                    currentMouseOver ~= TCS_SettingsMenu_Enable_Combat_Session_Display_Checkbox and
                    currentMouseOver ~= TCS_SettingsMenu_Enable_Combat_Session_Display_Checkbox.checkbox
                ) then
                    zo_callHandler( TCS_SettingsMenu_Enable_Combat_Session_Display_Checkbox, "OnMouseUp" )
                end

                self:SetCombatSessionHPSUpdateHidden()
            end,
        reference = "TCS_SettingsMenu_Show_HPS_Checkbox",
    })

    -- show total healing
    AddControl({
        type = "checkbox",
        name = "     " .. GetString( SI_TCS_SETTINGS_HEAL_TOTAL_SHOW_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_HEAL_TOTAL_TOOLTIP ),
        getFunc = function() return self.settings.totalHealing end,
        setFunc = function( _set )
                self.settings.totalHealing = _set

                local currentMouseOver = WINDOW_MANAGER:GetMouseOverControl()
                if( not self.settings.enableCombatSessionDisplay and
                    currentMouseOver ~= TCS_SettingsMenu_Enable_Combat_Session_Display_Checkbox and
                    currentMouseOver ~= TCS_SettingsMenu_Enable_Combat_Session_Display_Checkbox.checkbox
                ) then
                    zo_callHandler( TCS_SettingsMenu_Enable_Combat_Session_Display_Checkbox, "OnMouseUp" )
                end

                self:SetCombatSessionTotalHealingUpdateHidden()
            end,
        reference = "TCS_SettingsMenu_Total_Healing_Checkbox",
    })

    -- bg type
    AddControl({
        type = "dropdown",
        name = "     " .. GetString( SI_TCS_SETTINGS_BACKGROUND_TYPE_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_BACKGROUND_TYPE_TOOLTIP ),
        choices = {
                GetString( SI_TCS_SETTINGS_BACKGROUND_TYPE_A ),
                GetString( SI_TCS_SETTINGS_BACKGROUND_TYPE_B ),
                GetString( SI_TCS_SETTINGS_BACKGROUND_TYPE_NONE )
            },
        getFunc = function() return self.settings.combatSessionDisplayBGType end,
        setFunc = function( _type ) self.settings.combatSessionDisplayBGType = _type self:SetCombatSessionDisplayBGType( _type ) self:UpdateCombatSessionDisplays() end,
        reference = "TCS_SettingsMenu_BG_Type_Combat_Session_Display_Dropdown",
    })

    -- scale
    AddControl({
        type = "slider",
        name = "     " .. GetString( SI_TCS_SETTINGS_COMBAT_SESSION_DISPLAY_SCALE_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_COMBAT_SESSION_DISPLAY_SCALE_TOOLTIP ),
        min = 1,
        max = 6,
        getFunc = function() return self.settings.combatSessionDisplayScale end,
        setFunc = function( _set ) self.settings.combatSessionDisplayScale = _set self:UpdateCombatSessionDisplays() end,
        reference = "TCS_SettingsMenu_Scale_Combat_Session_Display_Slider",
    })

    -- bg transparency
    AddControl({
        type = "slider",
        name = "     " .. GetString( SI_TCS_SETTINGS_COMBAT_SESSION_DISPLAY_BG_TRANSPARENCY_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_COMBAT_SESSION_DISPLAY_BG_TRANSPARENCY_TOOLTIP ),
        min = 0,
        max = 100,
        getFunc = function() return self.settings.combatSessionDisplayBGTransparency end,
        setFunc = function( _set ) self.settings.combatSessionDisplayBGTransparency = _set self:SetCombatSessionDisplayBGTransparency( _set ) end,
        reference = "TCS_SettingsMenu_BG_Transparency_Combat_Session_Display_Slider",
    })

    -- text transparency
    AddControl({
        type = "slider",
        name = "     " .. GetString( SI_TCS_SETTINGS_COMBAT_SESSION_DISPLAY_TEXT_TRANSPARENCY_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_COMBAT_SESSION_DISPLAY_TEXT_TRANSPARENCY_TOOLTIP ),
        min = 0,
        max = 100,
        getFunc = function() return self.settings.combatSessionDisplayTextTransparency end,
        setFunc = function( _set ) self.settings.combatSessionDisplayTextTransparency = _set self:SetCombatSessionDisplayTextTransparency( _set ) end,
        reference = "TCS_SettingsMenu_Text_Transparency_Combat_Session_Display_Slider",
    })

    -- Header for the ability damage percentage
    AddControl({ type = "description", text = "" })
    AddControl({ type = "header", name = GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_LABEL ) })

    -- show bar indicators
    AddControl({
        type = "checkbox",
        name = GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_ENABLE_BAR_INDICATORS_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_ENABLE_BAR_INDICATORS_TOOLTIP ),
        getFunc = function() return self.settings.combatSessionAbilityDamageEnableBarIndicators end,
        setFunc = function( _set ) self.settings.combatSessionAbilityDamageEnableBarIndicators = _set self:UpdateAbilityDamagePercentagesEnabled() end,
        reference = "TCS_SettingsMenu_AbilityDamagePercentages_Enable_Bar_Indicators_Checkbox",
    })

    -- bar width
    AddControl({
        type = "slider",
        name = "     " .. GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_BAR_INDICATORS_WIDTH_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_BAR_INDICATORS_WIDTH_TOOLTIP ),
        min = 0,
        max = 100,
        getFunc = function() return self.settings.combatSessionAbilityDamageBarWidth end,
        setFunc = function( _set ) self.settings.combatSessionAbilityDamageBarWidth = _set self:UpdateAbilityDamagePercentagesWidth( true ) end,
        reference = "TCS_SettingsMenu_AbilityDamagePercentages_Bar_Indicators_Width_Slider",
    })

    -- bar alighment
    AddControl({
        type = "dropdown",
        name = "     " .. GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_BAR_INDICATORS_ALIGNMENT_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_BAR_INDICATORS_ALIGNMENT_TOOLTIP ),
        choices = {
                GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_BAR_ALIGNMENT_LEFT ),
                GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_BAR_ALIGNMENT_RIGHT )
            },
        getFunc = function() return self.settings.combatSessionAbilityDamageBarAlignment end,
        setFunc = function( _type ) self.settings.combatSessionAbilityDamageBarAlignment = _type self:UpdateAbilityDamagePercentagesAlignment() end,
        reference = "TCS_SettingsMenu_AbilityDamagePercentages_Bar_Indicators_Alignment_Dropdown",
    })

    -- bar color
    AddControl({
        type = "colorpicker",
        name = "     " .. GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_BAR_INDICATORS_COLOR_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_BAR_INDICATORS_COLOR_TOOLTIP ),
        getFunc = function() return self.settings.combatSessionAbilityDamageBarR, self.settings.combatSessionAbilityDamageBarG, self.settings.combatSessionAbilityDamageBarB end,
        setFunc = function( r, g, b )
                self.settings.combatSessionAbilityDamageBarR = r
                self.settings.combatSessionAbilityDamageBarG = g
                self.settings.combatSessionAbilityDamageBarB = b
                self:UpdateAbilityDamagePercentagesColor()
            end,
        reference = "TCS_SettingsMenu_AbilityDamagePercentages_Bar_Indicators_Color_Colorpicker",
    })

    -- show percentage indicator
    AddControl({
        type = "checkbox",
        name = GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_ENABLE_PERCENTAGE_INDICATORS_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_ENABLE_PERCENTAGE_INDICATORS_TOOLTIP ),
        getFunc = function() return self.settings.combatSessionAbilityDamageEnablePercentage end,
        setFunc = function( _set )
                self.settings.combatSessionAbilityDamageEnablePercentage = _set self:UpdateAbilityDamagePercentagesTextEnabled()
            end,
        reference = "TCS_SettingsMenu_AbilityDamagePercentages_Enable_Percentage_Indicators_Checkbox",
    })

    -- percentage size
    AddControl({
        type = "slider",
        name = "     " .. GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_PERCENTAGE_INDICATORS_SIZE_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_PERCENTAGE_INDICATORS_SIZE_TOOLTIP ),
        min = 1,
        max = 4,
        getFunc = function() return self.settings.combatSessionAbilityDamagePercentageSize end,
        setFunc = function( _set ) self.settings.combatSessionAbilityDamagePercentageSize = _set self:UpdateAbilityDamagePercentagesTextSize() end,
        reference = "TCS_SettingsMenu_AbilityDamagePercentages_Percentage_Indicators_Size_Slider",
    })

    -- percentage alignment
    AddControl({
        type = "dropdown",
        name = "     " .. GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_PERCENTAGE_INDICATORS_ALIGNMENT_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_PERCENTAGE_INDICATORS_ALIGNMENT_TOOLTIP ),
        choices = {
                GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_CENTER ),
                GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_TOP ),
                GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_BOTTOM ),
                GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_TOPRIGHT ),
                GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_TOPLEFT ),
                GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_BOTTOMRIGHT ),
                GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_BOTTOMLEFT ),
                GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_RIGHT ),
                GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_LEFT )
            },
        getFunc = function() return self.settings.combatSessionAbilityDamagePercentageAlignment end,
        setFunc = function( _type ) self.settings.combatSessionAbilityDamagePercentageAlignment = _type self:UpdateAbilityDamagePercentagesTextAlignment() end,
        reference = "TCS_SettingsMenu_AbilityDamagePercentages_Percentage_Indicators_Alignment_Dropdown",
    })

    AddControl({ type = "description", text = "" })
    AddControl({ type = "header", name = GetString( SI_TCS_SETTINGS_GLOBAL_LABEL ) })

    AddControl({
        type = "button",
        name = GetString( SI_TCS_SETTINGS_RESET_ALL_LABEL ),
        tooltip = GetString( SI_TCS_SETTINGS_RESET_ALL_TOOLTIP ),
        func = function() self:ResetAll() ReloadUI() end,
        warning = GetString( SI_TCS_SETTINGS_RESET ),
    })

end

function TCS:ResetAll()
    self.settings.abilityStatistics = {}

    self.settings.chatTabOptions = {}

    self.settings.abilityStatisticOptions.enableTooltip         = true
    self.settings.abilityStatisticOptions.showNormalAverage     = true
    self.settings.abilityStatisticOptions.showNormalMaximum     = true
    self.settings.abilityStatisticOptions.showNormalMinimum     = true
    self.settings.abilityStatisticOptions.showCriticalAverage   = true
    self.settings.abilityStatisticOptions.showCriticalMaximum   = true
    self.settings.abilityStatisticOptions.showCriticalMinimum   = true
    self.settings.abilityStatisticOptions.showTotalAverage      = true
    self.settings.abilityStatisticOptions.showCriticalPercent   = true

    self.settings.alliancePointsMinimum                 = 5
    self.settings.enableCombatSessionDisplay            = true
    self.settings.showDPS                               = true
    self.settings.totalDamage                           = true
    self.settings.showHPS                               = true
    self.settings.totalHealing                          = true
    self.settings.combatSessionDisplayBGType            = GetString( SI_TCS_SETTINGS_BACKGROUND_TYPE_A )
    self.settings.combatSessionDisplayScale             = 1
    self.settings.combatSessionDisplayBGTransparency    = 100
    self.settings.combatSessionDisplayTextTransparency  = 100

    self.settings.combatSessionDPSDisplayPoint          = BOTTOM
    self.settings.combatSessionDPSDisplayRelativeTo     = "GuiRoot"
    self.settings.combatSessionDPSDisplayRelativePoint  = BOTTOM
    self.settings.combatSessionDPSDisplayOffsetX        = -70
    self.settings.combatSessionDPSDisplayOffsetY        = -73

    self.settings.combatSessionTotalDamageDisplayPoint          = BOTTOM
    self.settings.combatSessionTotalDamageDisplayRelativeTo     = "GuiRoot"
    self.settings.combatSessionTotalDamageDisplayRelativePoint  = BOTTOM
    self.settings.combatSessionTotalDamageDisplayOffsetX        = -210
    self.settings.combatSessionTotalDamageDisplayOffsetY        = -73

    self.settings.combatSessionHPSDisplayPoint          = BOTTOM
    self.settings.combatSessionHPSDisplayRelativeTo     = "GuiRoot"
    self.settings.combatSessionHPSDisplayRelativePoint  = BOTTOM
    self.settings.combatSessionHPSDisplayOffsetX        = 70
    self.settings.combatSessionHPSDisplayOffsetY        = -73

    self.settings.combatSessionTotalHealingDisplayPoint         = BOTTOM
    self.settings.combatSessionTotalHealingDisplayRelativeTo    = "GuiRoot"
    self.settings.combatSessionTotalHealingDisplayRelativePoint = BOTTOM
    self.settings.combatSessionTotalHealingDisplayOffsetX       = 210
    self.settings.combatSessionTotalHealingDisplayOffsetY       = -73

    self.combatSessionExperienceGained     = 0
    self.combatSessionFinesseBonusExperiencePercentage  = 0

    self.combatSessionDamageDealt = 0
    self.combatSessionHealingDone = 0

    self:ChatCombatOutputInitializeLastExperiences()

    self:SetCombatSessionDPSUpdateHidden()
    self:SetCombatSessionTotalDamageUpdateHidden()
    self:SetCombatSessionHPSUpdateHidden()
    self:SetCombatSessionTotalHealingUpdateHidden()

    self.settings.combatSessionAbilityDamageBarAlignment            = GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_BAR_ALIGNMENT_LEFT )
    self.settings.combatSessionAbilityDamagePercentageAlignment     = GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_BOTTOMLEFT )
    self.settings.combatSessionAbilityDamageEnableBarIndicators     = true
    self.settings.combatSessionAbilityDamageBarWidth                = 15
    self.settings.combatSessionAbilityDamageBarR                    = 0.8039
    self.settings.combatSessionAbilityDamageBarG                    = 0.0862
    self.settings.combatSessionAbilityDamageBarB                    = 0.0862
    self.settings.combatSessionAbilityDamageEnablePercentage        = true
    self.settings.combatSessionAbilityDamagePercentageSize          = 1

    self:UpdateAbilityDamagePercentagesEnabled()
    self:UpdateAbilityDamagePercentagesWidth()
    self:UpdateAbilityDamagePercentagesColor()
    self:UpdateAbilityDamagePercentagesAlignment()
    self:UpdateAbilityDamagePercentagesTextEnabled()
    self:UpdateAbilityDamagePercentagesTextSize()
    self:UpdateAbilityDamagePercentagesTextAlignment()
    self:ResetAbilityDamagePercentage()
end

function TCS:OnUpdate()
    self:SkillTooltipOnUpdate()
end

----------------------------------
-- Chat Tab Options
----------------------------------
function TCS:ChatTabOptionsIntialize()
    -- Register handler
    ZO_PreHookHandler(
        ZO_ChatOptionsDialog, "OnShow",
        function()
            zo_callLater(
                function()
                    for i = 1, self.chatTabFilterCount do
                            local controlName   = "Filter" .. i
                            local settingName   = TCS_ChatOptionsDialogTsCombatStatsLabel[ controlName ].CheckButton.settingName
                            local containerId   = CHAT_OPTIONS.chatContainer.id
                            local tabIndex      = CHAT_OPTIONS:GetCurrentTabIndex()

                            if( self.settings.chatTabOptions[ containerId ] == nil ) then
                                self.settings.chatTabOptions[ containerId ] = {}
                            end

                            if( self.settings.chatTabOptions[ containerId ][ tabIndex ] == nil ) then
                                self.settings.chatTabOptions[ containerId ][ tabIndex ] = {}
                            end

                            ZO_CheckButton_SetCheckState(
                                TCS_ChatOptionsDialogTsCombatStatsLabel[ controlName ].CheckButton,
                                self.settings.chatTabOptions[ containerId ][ tabIndex ][ settingName ]
                            )
                    end
                end,
                10
            )
        end
    )

    -- Resize the chat tab options window
    ZO_ChatOptionsDialog:SetDimensionConstraints( 520, 915, 520, 915 )

    -- Create the new filter section
    TCS_ChatOptionsDialogTsCombatStatsLabel = WINDOW_MANAGER:CreateControl( "TCS_ChatOptionsDialogTsCombatStatsLabel", ZO_ChatOptionsDialog, CT_LABEL )
    TCS_ChatOptionsDialogTsCombatStatsLabel:SetText( "COMBAT STATISTICS" )
    TCS_ChatOptionsDialogTsCombatStatsLabel:SetFont( "ZoFontWinH3" )
    TCS_ChatOptionsDialogTsCombatStatsLabel:SetAnchor( TOPLEFT, ZO_ChatOptionsDialogGuildSection, BOTTOMLEFT, -10, 5 )

    -- Add the filters
    self:AddChatTabFilter( TCS_ChatOptionsDialogTsCombatStatsLabel, GetString( SI_TCS_FILTER_DAMAGE_DEALT ),           "damageDealt" )
    self:AddChatTabFilter( TCS_ChatOptionsDialogTsCombatStatsLabel, GetString( SI_TCS_FILTER_DAMAGE_RECEIVED ),        "damageReceived" )
    self:AddChatTabFilter( TCS_ChatOptionsDialogTsCombatStatsLabel, GetString( SI_TCS_FILTER_HEALING_DONE ),           "healingDone" )
    self:AddChatTabFilter( TCS_ChatOptionsDialogTsCombatStatsLabel, GetString( SI_TCS_FILTER_HEALING_RECEIVED ),       "healingReceived" )
    self:AddChatTabFilter( TCS_ChatOptionsDialogTsCombatStatsLabel, GetString( SI_TCS_FILTER_COMBAT_EXPERIENCE ),      "combatExperience" )
    self:AddChatTabFilter( TCS_ChatOptionsDialogTsCombatStatsLabel, GetString( SI_TCS_FILTER_FINESSE_EXPERIENCE ),     "finesseExperience" )
    self:AddChatTabFilter( TCS_ChatOptionsDialogTsCombatStatsLabel, GetString( SI_TCS_FILTER_SKILL_LINE_EXPERIENCE ),  "skillLineExperience" )
    self:AddChatTabFilter( TCS_ChatOptionsDialogTsCombatStatsLabel, GetString( SI_TCS_FILTER_ABILITY_EXPERIENCE ),     "abilityExperience" )
    self:AddChatTabFilter( TCS_ChatOptionsDialogTsCombatStatsLabel, GetString( SI_TCS_FILTER_BUFF_STATUS ),            "buffStatus" )
    self:AddChatTabFilter( TCS_ChatOptionsDialogTsCombatStatsLabel, GetString( SI_TCS_FILTER_DEBUFF_STATUS ),          "debuffStatus" )
    self:AddChatTabFilter( TCS_ChatOptionsDialogTsCombatStatsLabel, GetString( SI_TCS_FILTER_VETERAN_POINTS ),         "veteranPoints" )
    self:AddChatTabFilter( TCS_ChatOptionsDialogTsCombatStatsLabel, GetString( SI_TCS_FILTER_ALLIANCE_POINTS ),        "alliancePoints" )
    self:AddChatTabFilter( TCS_ChatOptionsDialogTsCombatStatsLabel, GetString( SI_TCS_FILTER_FINESSE_PROGRESSION ),    "finesseProgression" )
    self:AddChatTabFilter( TCS_ChatOptionsDialogTsCombatStatsLabel, GetString( SI_TCS_FILTER_COMBAT_START_END ),       "combatStartEnd" )
end

function TCS:AddChatTabFilter( _filterSection, _checkboxLabelText, _settingName )
    self.chatTabFilterCount = self.chatTabFilterCount + 1

    local controlName   = "Filter" .. self.chatTabFilterCount
    local offsetX       = self.chatTabFilterColumnCount * 240
    local offsetY       = self.chatTabFilterRowCount * 25

    TCS_ChatOptionsDialogTsCombatStatsLabel[ controlName ] = WINDOW_MANAGER:CreateControl( "TCS_ChatOptionsDialogTsCombatStats" .. controlName, _filterSection, CT_CONTROL )
    TCS_ChatOptionsDialogTsCombatStatsLabel[ controlName ]:SetAnchor( TOPLEFT, _filterSection, BOTTOMLEFT, 10 + offsetX, 12 + offsetY )

    TCS_ChatOptionsDialogTsCombatStatsLabel[ controlName ].CheckButton = CreateControlFromVirtual( "TCS_ChatOptionsDialogTsCombatStats" .. controlName .. "CheckButton", TCS_ChatOptionsDialogTsCombatStatsLabel[ controlName ], "ZO_CheckButton" )
    TCS_ChatOptionsDialogTsCombatStatsLabel[ controlName ].CheckButton:SetAnchor( TOPLEFT, TCS_ChatOptionsDialogTsCombatStatsLabel[ controlName ], BOTTOMLEFT, 0, 0 )
    TCS_ChatOptionsDialogTsCombatStatsLabel[ controlName ].CheckButton.settingName = _settingName

    ZO_CheckButton_SetToggleFunction(
        TCS_ChatOptionsDialogTsCombatStatsLabel[ controlName ].CheckButton,
        function( _button, _checked )
            local containerId = CHAT_OPTIONS.chatContainer.id
            local tabIndex = CHAT_OPTIONS:GetCurrentTabIndex()

            if( self.settings.chatTabOptions[ containerId ] == nil ) then
                self.settings.chatTabOptions[ containerId ] = {}
            end

            if( self.settings.chatTabOptions[ containerId ][ tabIndex ] == nil ) then
                self.settings.chatTabOptions[ containerId ][ tabIndex ] = {}
            end

            if( _checked ) then
                self.settings.chatTabOptions[ containerId ][ tabIndex ][ _settingName ] = _checked
            else
                self.settings.chatTabOptions[ containerId ][ tabIndex ][ _settingName ] = nil
            end
        end
    )

    TCS_ChatOptionsDialogTsCombatStatsLabel[ controlName ].CheckLabel = CreateControlFromVirtual( "TCS_ChatOptionsDialogTsCombatStats" .. controlName .. "CheckLabel", TCS_ChatOptionsDialogTsCombatStatsLabel[ controlName ], "ZO_ChatOptionsDefaultLabel" )
    TCS_ChatOptionsDialogTsCombatStatsLabel[ controlName ].CheckLabel:SetAnchor( TOPLEFT, TCS_ChatOptionsDialogTsCombatStatsLabel[ controlName ].CheckButton, TOPRIGHT, 5, -2 )
    TCS_ChatOptionsDialogTsCombatStatsLabel[ controlName ].CheckLabel:SetText( _checkboxLabelText )

    self.chatTabFilterColumnCount = self.chatTabFilterColumnCount + 1
    if( self.chatTabFilterColumnCount == 2 ) then
        self.chatTabFilterColumnCount = 0
        self.chatTabFilterRowCount = self.chatTabFilterRowCount + 1
    end
end

----------------------------------
-- Chat Combat Output
----------------------------------
function TCS:ChatCombatOutputInitialize()
    -- Initialize
    self:ChatCombatOutputInitializeLastExperiences()
    self:InitializeabilityInfoCache()
    self:InitializeAbilityNameProxies()

    -- Register Events
    self.control:RegisterForEvent( EVENT_PLAYER_COMBAT_STATE,               function( ... ) self:OnEventPlayerCombatState( ... )                        end )
    self.control:RegisterForEvent( EVENT_COMBAT_EVENT,                      function( ... ) self:OnEventCombatEvent( ... )                              end )
    self.control:RegisterForEvent( EVENT_SKILL_XP_UPDATE,                   function( ... ) self:OnEventSkillXPUpdate( ... )                            end )
    self.control:RegisterForEvent( EVENT_FINESSE_RANK_CHANGED,              function( ... ) self:OnEventFinesseRankChanged( ... )                       end )
    self.control:RegisterForEvent( EVENT_EFFECT_CHANGED,                    function( ... ) self:OnEventEffectChanged( ... )                            end )
    self.control:RegisterForEvent( EVENT_EXPERIENCE_UPDATE,                 function( ... ) self:OnEventExperienceUpdate( ... )                         end )
    self.control:RegisterForEvent( EVENT_VETERAN_POINTS_UPDATE,             function( ... ) self:OnEventVeteranPointsUpdate( ... )                      end )
    self.control:RegisterForEvent( EVENT_ALLIANCE_POINT_UPDATE,             function( ... ) self:OnEventAlliancePointUpdate( ... )                      end )
    self.control:RegisterForEvent( EVENT_SKILLS_FULL_UPDATE,                function( ... ) zo_callLater( self.InitializeAbilityNameProxies, 1000 ) zo_callLater( self.InitializeabilityInfoCache, 1100 ) zo_callLater( self.ChatCombatOutputInitializeLastExperiences, 1100 ) end )
end

function TCS:ChatCombatOutputInitializeLastExperiences()
    -- Player
    TCS.lastExperiences[ "player" ] = GetUnitXP( "player" )
    TCS.lastExperiences[ "playerVeteran" ] = GetUnitVeteranPoints( "player" )

    -- Skill Lines
    for i = 1, GetNumSkillTypes() do
        for j = 1, GetNumSkillLines( i ) do
            local skillLastRankXp, skillNextRankXP, skillCurrentXP = GetSkillLineXPInfo( i, j )
            local skillName, skillRank = GetSkillLineInfo( i, j )
            local skillNameIndex = zo_strformat( "<<z:1>>", skillName )

            if( skillCurrentXP == nil ) then skillCurrentXP = 0 end

            TCS.lastExperiences[ skillNameIndex ] = skillCurrentXP

            -- Abilities
            local abilityCount = GetNumSkillAbilities( i, j )

            for k = 1, abilityCount, 1 do
                local abilityName, abilityTexture, abilityEarnedRank, abilityPassive, abilityUltimate, abilityPurchased, abilityProgressionIndex = GetSkillAbilityInfo( i, j, k )
                local abilityLastRankXp, abilityNextRankXP, abilityCurrentXP, abilityAtMorph = GetAbilityProgressionXPInfo( abilityProgressionIndex )
                abilityNameIndex = zo_strformat( "<<z:1>>", abilityName )

                if( not abilityPassive and abilityPurchased ) then
                    local proxy = TCS.abilityNameProxies[ abilityNameIndex ]
                    if( proxy ~= nil ) then abilityNameIndex = proxy end

                    if( abilityCurrentXP == nil ) then abilityCurrentXP = 0 end

                    TCS.lastExperiences[ abilityNameIndex ] = abilityCurrentXP
                end
            end
        end
    end
end

function TCS:InitializeabilityInfoCache()
    for i = 1, GetNumSkillTypes() do
        if( TCS.abilityInfoCache[ i ] == nil ) then TCS.abilityInfoCache[ i ] = {} end

        for j = 1, GetNumSkillLines( i ) do
            if( TCS.abilityInfoCache[ i ][ j ] == nil ) then TCS.abilityInfoCache[ i ][ j ] = { name = "", abilityCount = 0, abilities = {} } end

            local skillName, skillRank = GetSkillLineInfo( i, j )
            TCS.abilityInfoCache[ i ][ j ].name = zo_strformat( "<<z:1>>", skillName )

            local abilityCount = GetNumSkillAbilities( i, j )
            TCS.abilityInfoCache[ i ][ j ].abilityCount = abilityCount

            for k = 1, abilityCount do
                if( TCS.abilityInfoCache[ i ][ j ].abilities[ k ] == nil ) then TCS.abilityInfoCache[ i ][ j ].abilities[ k ] = { name = "", progressionIndex = 0 } end

                local abilityName, abilityTexture, abilityEarnedRank, abilityPassive, abilityUltimate, abilityPurchased, abilityProgressionIndex = GetSkillAbilityInfo( i, j, k )
                local progressionName, progressionMorph, progressionRank = GetAbilityProgressionInfo( abilityProgressionIndex )
                abilityName = zo_strformat( "<<z:1>>", abilityName )

                if( not abilityPassive ) then
                    local proxy = TCS.abilityNameProxies[ abilityName ]
                    if( proxy ~= nil ) then abilityName = proxy end

                    TCS.abilityInfoCache[ i ][ j ].abilities[ k ].name = abilityName
                    TCS.abilityInfoCache[ i ][ j ].abilities[ k ].progressionIndex = abilityProgressionIndex
                    TCS.abilityInfoCache[ i ][ j ].abilities[ k ].isPassive = abilityPassive
                    TCS.abilityInfoCache[ i ][ j ].abilities[ k ].isPurchased = abilityPurchased
                end
            end
        end
    end
end

--_tcsDebugAbilityProxies = nil
function TCS:InitializeAbilityNameProxies()
    for i = 1, GetNumSkillTypes() do
        for j = 1, GetNumSkillLines( i ) do
            local abilityCount = GetNumSkillAbilities( i, j )

            for k = 1, abilityCount, 1 do
                local abilityName, abilityTexture, abilityEarnedRank, abilityPassive, abilityUltimate, abilityPurchased, abilityProgressionIndex = GetSkillAbilityInfo( i, j, k )

                if( abilityProgressionIndex ~= nil ) then
                    local progressionName, progressionMorph, progressionRank = GetAbilityProgressionInfo( abilityProgressionIndex )

                    abilityName = zo_strformat( "<<z:1>>", abilityName )
                    progressionName = zo_strformat( "<<z:1>>", progressionName )

                    TCS.abilityNameProxies[ abilityName ] = abilityName
                    TCS.abilityNameProxies[ progressionName ] = abilityName
                end
            end
        end
    end

    -- destruction staff, element specific names
    if( TCS_LOCALE == "EN" ) then
        TCS.abilityNameProxies[ "fire touch" ] = "destructive touch"
        TCS.abilityNameProxies[ "frost touch" ] = "destructive touch"
        TCS.abilityNameProxies[ "shock touch" ] = "destructive touch"

        TCS.abilityNameProxies[ "fire clench" ] = "destructive clench"
        TCS.abilityNameProxies[ "frost clench" ] = "destructive clench"
        TCS.abilityNameProxies[ "shock clench" ] = "destructive clench"

        TCS.abilityNameProxies[ "fire reach" ] = "destructive reach"
        TCS.abilityNameProxies[ "frost reach" ] = "destructive reach"
        TCS.abilityNameProxies[ "shock reach" ] = "destructive reach"

        TCS.abilityNameProxies[ "wall of fire" ] = "wall of elements"
        TCS.abilityNameProxies[ "wall of frost" ] = "wall of elements"
        TCS.abilityNameProxies[ "wall of storms" ] = "wall of elements"

        TCS.abilityNameProxies[ "unstable wall of fire" ] = "usnstable wall of elements"
        TCS.abilityNameProxies[ "unstable wall of frost" ] = "unstable wall of elements"
        TCS.abilityNameProxies[ "unstable wall of storms" ] = "unstable wall of elements"

        TCS.abilityNameProxies[ "blockade of fire" ] = "elemental blockade"
        TCS.abilityNameProxies[ "blockade of frost" ] = "elemental blockade"
        TCS.abilityNameProxies[ "blockade of storms" ] = "elemental blockade"

        TCS.abilityNameProxies[ "fire impulse" ] = "impulse"
        TCS.abilityNameProxies[ "frost impulse" ] = "impulse"
        TCS.abilityNameProxies[ "shock impulse" ] = "impulse"

        TCS.abilityNameProxies[ "ring of fire" ] = "elemental ring"
        TCS.abilityNameProxies[ "ring of frost" ] = "elemental ring"
        TCS.abilityNameProxies[ "ring of shock" ] = "elemental ring"

        TCS.abilityNameProxies[ "fiery pulsar" ] = "pulsar"
        TCS.abilityNameProxies[ "icy pulsar" ] = "pulsar"
        TCS.abilityNameProxies[ "electric pulsar" ] = "pulsar"

        TCS.abilityNameProxies[ "overload light attack" ] = "overload"
        TCS.abilityNameProxies[ "overload heavy attack" ] = "overload"
        TCS.abilityNameProxies[ "power overload heavy attack" ] = "overload"
    elseif( TCS_LOCALE == "FR" ) then
        TCS.abilityNameProxies[ "toucher de feu" ] = "toucher destructeur"
        TCS.abilityNameProxies[ "toucher de givre" ] = "toucher destructeur"
        TCS.abilityNameProxies[ "toucher de foudre" ] = "toucher destructeur"

        TCS.abilityNameProxies[ "prise de feu" ] = "prise destructrice"
        TCS.abilityNameProxies[ "prise de givre" ] = "prise destructrice"
        TCS.abilityNameProxies[ "prise de foudre" ] = "prise destructrice"

        TCS.abilityNameProxies[ "allonge de feu" ] = "allonge destructrice"
        TCS.abilityNameProxies[ "allonge de givre" ] = "allonge destructrice"
        TCS.abilityNameProxies[ "allonge de foudre" ] = "allonge destructrice"

        TCS.abilityNameProxies[ "mur de feu" ] = "mur élémentaire"
        TCS.abilityNameProxies[ "mur de givre" ] = "mur élémentaire"
        TCS.abilityNameProxies[ "mur de tempêtes" ] = "mur élémentaire"

        TCS.abilityNameProxies[ "mur de feu instable" ] = "mur élémentaire instable"
        TCS.abilityNameProxies[ "mur de givre instable" ] = "mur élémentaire instable"
        TCS.abilityNameProxies[ "mur de tempêtes instable" ] = "mur élémentaire instable"

        TCS.abilityNameProxies[ "barrage de feu" ] = "rempart élémentaire"
        TCS.abilityNameProxies[ "barrage de givre" ] = "rempart élémentaire"
        TCS.abilityNameProxies[ "barrage de tempêtes" ] = "rempart élémentaire"

        TCS.abilityNameProxies[ "impulsion de feu" ] = "impulsion"
        TCS.abilityNameProxies[ "impulsion de froid" ] = "impulsion"
        TCS.abilityNameProxies[ "impulsion de foundre" ] = "impulsion"

        TCS.abilityNameProxies[ "fire ring" ] = "cercle élémentaire"
        TCS.abilityNameProxies[ "cercle de givre" ] = "cercle élémentaire"
        TCS.abilityNameProxies[ "cercle de foudre" ] = "cercle élémentaire"

        TCS.abilityNameProxies[ "pulsar embrasé" ] = "pulsar"
        TCS.abilityNameProxies[ "pulsar de glace" ] = "pulsar"
        TCS.abilityNameProxies[ "pulsar électrique" ] = "pulsar"

        TCS.abilityNameProxies[ "décharge" ] = "surcharge"
        TCS.abilityNameProxies[ "attaque lourde surcharge" ] = "surcharge"
        TCS.abilityNameProxies[ "attaque lourde surcharge puissante" ] = "surcharge"
    elseif( TCS_LOCALE == "DE" ) then
        TCS.abilityNameProxies[ "feuerberührung" ] = "zerstörerische berührung"
        TCS.abilityNameProxies[ "frostberührung" ] = "zerstörerische berührung"
        TCS.abilityNameProxies[ "schockberührung" ] = "zerstörerische berührung"

        TCS.abilityNameProxies[ "feuerhieb" ] = "zerstörerischer hieb"
        TCS.abilityNameProxies[ "frosthieb" ] = "zerstörerischer hieb"
        TCS.abilityNameProxies[ "schockhieb" ] = "zerstörerischer hieb"

        TCS.abilityNameProxies[ "feuerfaust" ] = "zerstörerische faust"
        TCS.abilityNameProxies[ "frostfaust" ] = "zerstörerische faust"
        TCS.abilityNameProxies[ "schockfaust" ] = "zerstörerische faust"

        TCS.abilityNameProxies[ "feuerwand" ] = "elementare wand"
        TCS.abilityNameProxies[ "frostwand" ] = "elementare wand"
        TCS.abilityNameProxies[ "sturmwand" ] = "elementare wand"

        TCS.abilityNameProxies[ "instabile feuerwand" ] = "instabile elementare wand"
        TCS.abilityNameProxies[ "instabile frostwand" ] = "instabile elementare wand"
        TCS.abilityNameProxies[ "instabile sturmwand" ] = "instabile elementare wand"

        TCS.abilityNameProxies[ "feuerblockade" ] = "elementare blockade"
        TCS.abilityNameProxies[ "frostblockade" ] = "elementare blockade"
        TCS.abilityNameProxies[ "sturmblockade" ] = "elementare blockade"

        TCS.abilityNameProxies[ "feuerimpuls" ] = "impuls"
        TCS.abilityNameProxies[ "frostimpuls" ] = "impuls"
        TCS.abilityNameProxies[ "schockimpuls" ] = "impuls"

        TCS.abilityNameProxies[ "feuerring" ] = "elementarer ring"
        TCS.abilityNameProxies[ "frostring" ] = "elementarer ring"
        TCS.abilityNameProxies[ "schockring" ] = "elementarer ring"

        TCS.abilityNameProxies[ "feuriger pulsar" ] = "pulsar"
        TCS.abilityNameProxies[ "eisiger pulsar" ] = "pulsar"
        TCS.abilityNameProxies[ "elektrisierender pulsar" ] = "pulsar"

        TCS.abilityNameProxies[ "geladener schlag" ] = "überladung"
        TCS.abilityNameProxies[ "geladener hieb" ] = "überladung"
        TCS.abilityNameProxies[ "kräftiger geladener hieb" ] = "überladung"
    end

    -- typos
    if( TCS_LOCALE == "EN" ) then
        TCS.abilityNameProxies[ "critcal charge" ] = TCS.abilityNameProxies[ "critical charge" ]
    elseif( TCS_LOCALE == "FR" ) then
        --TCS.abilityNameProxies[ "critcal charge" ] = TCS.abilityNameProxies[ "critical charge" ]
    elseif( TCS_LOCALE == "DE" ) then
        --TCS.abilityNameProxies[ "critcal charge" ] = TCS.abilityNameProxies[ "critical charge" ]
    end

    --_tcsDebugAbilityProxies = TCS.abilityNameProxies
end

function TCS:InitializeAbilityStatistic( _abilityNameIndex )
    self.settings.abilityStatistics[ _abilityNameIndex ] = {}

    self.settings.abilityStatistics[ _abilityNameIndex ].normalCount        = 0
    self.settings.abilityStatistics[ _abilityNameIndex ].normalLowest       = 0
    self.settings.abilityStatistics[ _abilityNameIndex ].normalHighest      = 0
    self.settings.abilityStatistics[ _abilityNameIndex ].normalAverage      = 0
    self.settings.abilityStatistics[ _abilityNameIndex ].criticalCount      = 0
    self.settings.abilityStatistics[ _abilityNameIndex ].criticalLowest     = 0
    self.settings.abilityStatistics[ _abilityNameIndex ].criticalHighest    = 0
    self.settings.abilityStatistics[ _abilityNameIndex ].criticalAverage    = 0
    self.settings.abilityStatistics[ _abilityNameIndex ].criticalPercent    = 0.0
    self.settings.abilityStatistics[ _abilityNameIndex ].damageType         = ""
end

function TCS:OnEventVeteranPointsUpdate( _eventCode, _unitTag, _currentPoints, _maxPoints, _reason )
    if _unitTag == "player" then
        local veteranPointsGained = _currentPoints - self.lastExperiences[ "playerVeteran" ]
        self.lastExperiences[ "playerVeteran" ] = _currentPoints

        self:DisplayMessage( "veteranPoints", SI_TCS_VETERAN_POINTS, veteranPointsGained )
    end
end

function TCS:OnEventExperienceUpdate( _eventCode, _unitTag, _currentExp, _maxExp, _reason)
    if _unitTag == "player" then
        local expGained = _currentExp - self.lastExperiences[ "player" ]
        self.combatSessionExperienceGained = self.combatSessionExperienceGained + expGained 
        self.lastExperiences[ "player" ] = _currentExp
        
        self:DisplayMessage( "combatExperience", SI_TCS_EXPERIENCE_POINTS, expGained )
    end 
end

function TCS:OnEventAlliancePointUpdate( _eventCode, _alliancePoints, _playSound, _difference )
    if( _difference > self.settings.alliancePointsMinimum ) then
        self:DisplayMessage( "alliancePoints", SI_TCS_ALLIANCE_POINTS, _difference )
    end
end

function TCS:OnEventCombatEvent( _eventCode, _result, _isError, _abilityName, _abilityGraphic, _abilityActionSlotType, _sourceName, _sourceType, _targetName, _targetType, _hitValue, _powerType, _damageType, _log )
    if( _hitValue < 1 ) then return end

    local playerName        = zo_strformat( "<<C:1>>", GetUnitName( "player" ) )
    local sourceName        = zo_strformat( "<<C:1>>", _sourceName )
    local targetName        = zo_strformat( "<<C:1>>", _targetName )
    local abilityName       = zo_strformat( "<<t:1>>", _abilityName )
    local abilityNameIndex  = zo_strformat( "<<z:1>>", _abilityName )
    local proxyName         = self.abilityNameProxies[ abilityNameIndex ]

    if( proxyName ~= nil ) then
        abilityName = zo_strformat( "<<t:1>>", proxyName )
        abilityNameIndex = proxyName
    end

    if( _result == ACTION_RESULT_BLOCKED_DAMAGE ) then
        -- Damage Blocked
        if( targetName == playerName and sourceName ~= playerName ) then
            self:DisplayMessage( "damageReceived", SI_TCS_DAMAGE_BLOCKED, _hitValue )
        end
    elseif( _result == ACTION_RESULT_FALL_DAMAGE ) then
        -- Fall Damage-- Fall Damage
        if( targetName == playerName and sourceName ~= playerName ) then
            self:DisplayMessage( "damageReceived", SI_TCS_DAMAGE_FALL, _hitValue )
        end
    end

    -- Ability Statistics
    if( sourceName == playerName ) then

        if( _result == ACTION_RESULT_DAMAGE             or _result == ACTION_RESULT_DOT_TICK            or _result == ACTION_RESULT_HEAL            or _result == ACTION_RESULT_HOT_TICK or
            _result == ACTION_RESULT_CRITICAL_DAMAGE    or _result == ACTION_RESULT_DOT_TICK_CRITICAL   or _result == ACTION_RESULT_CRITICAL_HEAL   or _result == ACTION_RESULT_HOT_TICK_CRITICAL
        ) then

        if( #abilityNameIndex > 0 and self.settings.abilityStatistics[ abilityNameIndex ] == nil ) then
                self:InitializeAbilityStatistic( abilityNameIndex )
            end
        end

        if( self.settings.abilityStatistics[ abilityNameIndex ] ~= nil ) then
            if( self.settings.abilityStatistics[ abilityNameIndex ].normalCount     > 1000 ) then
                self.settings.abilityStatistics[ abilityNameIndex ].normalCount     = 10
                self.settings.abilityStatistics[ abilityNameIndex ].criticalCount   = 10 * ( self.settings.abilityStatistics[ abilityNameIndex ].criticalPercent / 100 )
            end

            if( _result == ACTION_RESULT_DAMAGE or _result == ACTION_RESULT_DOT_TICK or _result == ACTION_RESULT_HEAL or _result == ACTION_RESULT_HOT_TICK ) then
                if( _hitValue < self.settings.abilityStatistics[ abilityNameIndex ].normalLowest ) then
                    self.settings.abilityStatistics[ abilityNameIndex ].normalLowest = _hitValue

                    if( self.settings.abilityStatistics[ abilityNameIndex ].normalHighest == 0 ) then
                        self.settings.abilityStatistics[ abilityNameIndex ].normalHighest = _hitValue
                    end
                elseif( _hitValue > self.settings.abilityStatistics[ abilityNameIndex ].normalHighest ) then
                    self.settings.abilityStatistics[ abilityNameIndex ].normalHighest = _hitValue

                    if( self.settings.abilityStatistics[ abilityNameIndex ].normalLowest == 0 ) then
                        self.settings.abilityStatistics[ abilityNameIndex ].normalLowest = _hitValue
                    end
                end

                if( self.settings.abilityStatistics[ abilityNameIndex ].normalAverage == 0 ) then
                    self.settings.abilityStatistics[ abilityNameIndex ].normalAverage = _hitValue
                end

                self.settings.abilityStatistics[ abilityNameIndex ].normalCount     = self.settings.abilityStatistics[ abilityNameIndex ].normalCount + 1
                self.settings.abilityStatistics[ abilityNameIndex ].normalAverage   = ( _hitValue + self.settings.abilityStatistics[ abilityNameIndex ].normalAverage ) / 2
            elseif( _result == ACTION_RESULT_CRITICAL_DAMAGE or _result == ACTION_RESULT_DOT_TICK_CRITICAL or _result == ACTION_RESULT_CRITICAL_HEAL or _result == ACTION_RESULT_HOT_TICK_CRITICAL ) then
                if( _hitValue < self.settings.abilityStatistics[ abilityNameIndex ].criticalLowest ) then
                    self.settings.abilityStatistics[ abilityNameIndex ].criticalLowest = _hitValue

                    if( self.settings.abilityStatistics[ abilityNameIndex ].criticalHighest == 0 ) then
                        self.settings.abilityStatistics[ abilityNameIndex ].criticalHighest = _hitValue
                    end
                elseif( _hitValue > self.settings.abilityStatistics[ abilityNameIndex ].criticalHighest ) then
                    self.settings.abilityStatistics[ abilityNameIndex ].criticalHighest = _hitValue

                    if( self.settings.abilityStatistics[ abilityNameIndex ].criticalLowest == 0 ) then
                        self.settings.abilityStatistics[ abilityNameIndex ].criticalLowest = _hitValue
                    end
                end

                if( self.settings.abilityStatistics[ abilityNameIndex ].criticalAverage == 0 ) then
                    self.settings.abilityStatistics[ abilityNameIndex ].criticalAverage = _hitValue
                end

                self.settings.abilityStatistics[ abilityNameIndex ].criticalCount   = self.settings.abilityStatistics[ abilityNameIndex ].criticalCount + 1
                self.settings.abilityStatistics[ abilityNameIndex ].criticalAverage = ( _hitValue + self.settings.abilityStatistics[ abilityNameIndex ].criticalAverage ) / 2
                self.settings.abilityStatistics[ abilityNameIndex ].criticalPercent = ( self.settings.abilityStatistics[ abilityNameIndex ].criticalCount / ( self.settings.abilityStatistics[ abilityNameIndex ].normalCount + self.settings.abilityStatistics[ abilityNameIndex ].criticalCount ) ) * 100
            end

            if( _result == ACTION_RESULT_DAMAGE or _result == ACTION_RESULT_DOT_TICK or _result == ACTION_RESULT_CRITICAL_DAMAGE or _result == ACTION_RESULT_DOT_TICK_CRITICAL ) then
                self.settings.abilityStatistics[ abilityNameIndex ].damageType =  zo_strformat( GetString( SI_TCS_TYPE_DAMAGE ), GetString( _G[ self.damageTypeStrings[ _damageType ] ] ) )

                self.combatSessionDamageDealt = self.combatSessionDamageDealt + _hitValue
                self:UpdateCombatSessionDisplays()
                self:UpdateAbilityDamagePercentage( zo_strformat( "<<z:1>>", abilityName ), _hitValue )
            elseif( _result == ACTION_RESULT_HEAL or _result == ACTION_RESULT_HOT_TICK or _result == ACTION_RESULT_CRITICAL_HEAL or _result == ACTION_RESULT_HOT_TICK_CRITICAL) then
                self.settings.abilityStatistics[ abilityNameIndex ].damageType = GetString( SI_TCS_TYPE_HEALING )

                self.combatSessionHealingDone = self.combatSessionHealingDone + _hitValue
                self:UpdateCombatSessionDisplays()
            end
        end
    end

    -- Damage Dealt / Taken
    if( _result == ACTION_RESULT_DAMAGE or _result == ACTION_RESULT_DOT_TICK ) then
        if( targetName ~= playerName and sourceName == playerName ) then
            self:DisplayMessage( "damageDealt", SI_TCS_DAMAGE_DEALT, _hitValue, GetString( _G[ self.damageTypeStrings[ _damageType ] ] ), abilityName, targetName )
        elseif( targetName == playerName ) then
            if( sourceName ~= playerName ) then
                self:DisplayMessage( "damageReceived", SI_TCS_DAMAGE_TAKEN, _hitValue )
            elseif( sourceName == playerName ) then
                self:DisplayMessage( "damageReceived", SI_TCS_DAMAGE_TAKEN_SELF, _hitValue, GetString( _G[ self.damageTypeStrings[ _damageType ] ] ), abilityName )
            end
        end
    elseif( _result == ACTION_RESULT_CRITICAL_DAMAGE or _result == ACTION_RESULT_DOT_TICK_CRITICAL ) then
        if( targetName ~= playerName and sourceName == playerName ) then
            self:DisplayMessage( "damageDealt", SI_TCS_DAMAGE_DEALT_CRITICAL, _hitValue, GetString( _G[ self.damageTypeStrings[ _damageType ] ] ), abilityName, targetName )
        elseif( targetName == playerName ) then
            if( sourceName ~= playerName ) then
                self:DisplayMessage( "damageReceived", SI_TCS_DAMAGE_TAKEN_CRITICAL, _hitValue )
            elseif( sourceName == playerName ) then
                self:DisplayMessage( "damageReceived", SI_TCS_DAMAGE_TAKEN_CRITICAL_SELF, _hitValue, GetString( _G[ self.damageTypeStrings[ _damageType ] ] ), abilityName )
            end
        end
    end

    -- Healing Done/Recieved
    if( _result == ACTION_RESULT_HEAL or _result == ACTION_RESULT_HOT_TICK ) then
        if( sourceName == playerName and targetName ~= playerName ) then
            self:DisplayMessage( "healingDone", SI_TCS_HEALING_DONE, targetName, _hitValue, abilityName )
        elseif( targetName == playerName ) then
            if( sourceName ~= playerName ) then
                self:DisplayMessage( "healingReceived", SI_TCS_HEALING_RECEIVED, _hitValue )
            elseif( sourceName == playerName ) then
                self:DisplayMessage( "healingReceived", SI_TCS_HEALING_RECEIVED_SELF, _hitValue, abilityName )
            end
        end
    elseif( _result == ACTION_RESULT_CRITICAL_HEAL or _result == ACTION_RESULT_HOT_TICK_CRITICAL ) then
        if( sourceName == playerName and targetName ~= playerName ) then
            self:DisplayMessage( "healingDone", SI_TCS_HEALING_DONE_CRITICAL, targetName, _hitValue, abilityName )
        elseif( targetName == playerName ) then
            if( sourceName ~= playerName ) then
                self:DisplayMessage( "healingReceived", SI_TCS_HEALING_RECEIVED_CRITICAL, _hitValue )
            elseif( sourceName == playerName ) then
                self:DisplayMessage( "healingReceived", SI_TCS_HEALING_RECEIVED_CRITICAL_SELF, _hitValue, abilityName )
            end
        end
    end

    if( _result == ACTION_RESULT_DIED_XP and sourceName == playerName and not IsUnitVeteran( "player" ) ) then
        -- Kill Experience
        self.combatSessionExperienceGained = self.combatSessionExperienceGained + _hitValue

        self:DisplayMessage( "combatExperience", SI_TCS_COMBAT_EXPERIENCE, _hitValue, _targetName )
    elseif( _result == ACTION_RESULT_DIED and sourceName == playerName and not IsUnitVeteran( "player" ) ) then
        -- Kill No Experience
        self:DisplayMessage( "combatExperience", SI_TCS_COMBAT_NO_EXPERIENCE, _targetName )
    end
end

function TCS:OnEventEffectChanged( _eventCode, _changeType, _effectSlot, _effectName, _unitTag, _beginTime, _endTime, _stackCount, _iconName, _buffType, _effectType, _abilityType, _statusEffectType )
    if( _unitTag ~= "player" or effectName == nil or #effectName == 0 ) then
        return
    end

    if( _effectType == BUFF_EFFECT_TYPE_BUFF ) then
        if( _changeType == 1 ) then
            self:DisplayMessage( "buffStatus", SI_TCS_BUFF_STATUS_GAIN,  _effectName )
        elseif( _changeType == 2 ) then
            self:DisplayMessage( "buffStatus", SI_TCS_BUFF_STATUS_LOOSE, _effectName )
        end
    elseif( _effectType == BUFF_EFFECT_TYPE_DEBUFF ) then
        if( _changeType == 1 ) then
            self:DisplayMessage( "debuffStatus", SI_TCS_DEBUFF_STATUS_GAIN,  _effectName )
        elseif( _changeType == 2 ) then
            self:DisplayMessage( "debuffStatus", SI_TCS_DEBUFF_STATUS_LOOSE, _effectName )
        end
    else
        if( _changeType == 1 ) then
            self:DisplayMessage( "buffStatus", SI_TCS_EFFECT_STATUS_GAIN,    _effectName )
        elseif( _changeType == 2 ) then
            self:DisplayMessage( "buffStatus", SI_TCS_EFFECT_STATUS_LOOSE,   _effectName )
        end
    end
end

function TCS:OnEventPlayerCombatState( _eventCode, _inCombat )
    self.isInCombat = _inCombat

    if( _inCombat ) then
        self.combatSessionExperienceGained     = 0
        self.combatSessionFinesseBonusExperiencePercentage  = 0
        self.lastExperiences[ "player" ]        = GetUnitXP( 'Player' )

        if( IsUnitVeteran( "player" ) ) then
            self.lastExperiences[ "playerVeteran" ] = GetUnitVeteranPoints( "player" )
        end

        self.combatSessionStartTime                     = GetGameTimeMilliseconds()
        self.combatSessionDamageDealt                   = 0
        self.combatSessionHealingDone                   = 0
        self.combatSessionDPSNextSampleUpdate           = 0
        self.combatSessionAverageDPSSampleCounter       = 0
        self.combatSessionAverageDPSSamples             = {}
        self.combatSessionHPSNextSampleUpdate           = 0
        self.combatSessionAverageHPSSampleCounter       = 0
        self.combatSessionAverageHPSSamples             = {}
        self.combatSessionSkillAbilityExperienceGained  = {}
        self:ResetAbilityDamagePercentage()

        self:DisplayMessage( "combatStartEnd", SI_TCS_COMBAT_START )
    else
        zo_callLater(
            function()
                for i, v in pairs( self.combatSessionSkillAbilityExperienceGained ) do
                    if( v.skillLineExperienceGained ~= 0 ) then self:DisplayMessage( "skillLineExperience", SI_TCS_SKILL_LINE_EXPERIENCE, v.skillLineExperienceGained, i ) end
                    if( v.abilitiesExperienceGained ~= 0 ) then self:DisplayMessage( "abilityExperience", SI_TCS_ABILITY_EXPERIENCE, v.abilitiesExperienceGained, i ) end
                end

                if( self.combatSessionFinesseBonusExperiencePercentage > 0 ) then
                    local finesseExperience = zo_round( self.combatSessionExperienceGained * self.combatSessionFinesseBonusExperiencePercentage )

                    self:DisplayMessage( "finesseExperience", SI_TCS_FINESSE_EXPERIENCE, finesseExperience )
                end

                self:DisplayMessage( "combatStartEnd", SI_TCS_COMBAT_END )
            end,
            50
        )
    end
end

function TCS:OnEventFinesseRankChanged( _eventCode, _unitTag, _rankNum, _name, _xpBonus, _loot )
    if( _unitTag ~= "player" ) then return end

    self.combatSessionFinesseBonusExperiencePercentage = _xpBonus / 100

    if( _loot ) then
        if( IsUnitVeteran( "player" ) ) then
            self:DisplayMessage( "finesseProgression", SI_TCS_FINESSE_PROGRESSION_LOOT_NOXP, _name )
        else
            self:DisplayMessage( "finesseProgression", SI_TCS_FINESSE_PROGRESSION_LOOT, _name, _xpBonus )
        end
    else
        if( IsUnitVeteran( "player" ) ) then
            self:DisplayMessage( "finesseProgression", SI_TCS_FINESSE_PROGRESSION_NOXP, _name )
        else
            self:DisplayMessage( "finesseProgression", SI_TCS_FINESSE_PROGRESSION, _name, _xpBonus )
        end
    end
end

function TCS:OnEventSkillXPUpdate( _eventCode, _skillType, _skillIndex, _lastRankXP, _nextRankXP, _currentXP )
    -- Skill Experience
    -- Calculate the skill experience gained
    local skillName = self.abilityInfoCache[ _skillType ][ _skillIndex ].name
    local skillNameIndex = zo_strformat( "<<z:1>>", skillName )

    if( skillType == SKILL_TYPE_RACIAL ) then return end --I have no idea why

    local skillExperienceGained = _currentXP - self.lastExperiences[ skillNameIndex ]
    self.lastExperiences[ skillNameIndex ] = _currentXP

    if( skillExperienceGained ~= 0 ) then
        if( self.combatSessionSkillAbilityExperienceGained[ skillName ] == nil ) then
            self.combatSessionSkillAbilityExperienceGained[ skillName ] =
            {
                skillLineExperienceGained = 0,
                abilitiesExperienceGained = 0
            }
        end

        self.combatSessionSkillAbilityExperienceGained[ skillName ].skillLineExperienceGained = self.combatSessionSkillAbilityExperienceGained[ skillName ].skillLineExperienceGained + skillExperienceGained
    end

    -- Ability Experience
    local abilityCount = self.abilityInfoCache[ _skillType ][ _skillIndex ].abilityCount

    for i = 1, abilityCount do
        local abilityIsPurchased = self.abilityInfoCache[ _skillType ][ _skillIndex ].abilities[ i ].isPurchased
        local abilityIsPassive = self.abilityInfoCache[ _skillType ][ _skillIndex ].abilities[ i ].isPassive

        if( abilityIsPurchased and not abilityIsPassive ) then
            local abilityName = self.abilityInfoCache[ _skillType ][ _skillIndex ].abilities[ i ].name
            local abilityNameIndex = zo_strformat( "<<z:1>>", abilityName )
            local abilityLastRankXp, abilityNextRankXP, abilityCurrentXP, abilityAtMorph = GetAbilityProgressionXPInfo( self.abilityInfoCache[ _skillType ][ _skillIndex ].abilities[ i ].progressionIndex )

            -- Calculate the ability experience gained
            if ( self.lastExperiences[ abilityNameIndex ] == nil ) then self.lastExperiences[ abilityNameIndex ] = 0 end

            local abilityExperienceGained = abilityCurrentXP - self.lastExperiences[ abilityNameIndex ]
            self.lastExperiences[ abilityNameIndex ] = abilityCurrentXP

            if( abilityExperienceGained ~= 0 ) then
                if( self.combatSessionSkillAbilityExperienceGained[ skillName ] == nil ) then
                    self.combatSessionSkillAbilityExperienceGained[ skillName ] =
                    {
                        skillLineExperienceGained = 0,
                        abilitiesExperienceGained = 0
                    }
                end

                self.combatSessionSkillAbilityExperienceGained[ skillName ].abilitiesExperienceGained = self.combatSessionSkillAbilityExperienceGained[ skillName ].abilitiesExperienceGained + abilityExperienceGained

                break
            end
        end
    end
end

function TCS:DisplayMessage( _filter, _message, ... )
    for i, v in pairs(self.settings.chatTabOptions) do
        for j in pairs(v) do
            if( v[ j ][ _filter ] == true ) then
                if CHAT_SYSTEM.containers[ i ] and CHAT_SYSTEM.containers[ i ].windows[ j ] then
                    local tab = CHAT_SYSTEM.containers[ i ].windows[ j ].buffer
                    if( tab == nil ) then return end
                    tab:AddMessage( zo_strformat( GetString( _message ), ... ) )
                else
                    self.settings.chatTabOptions[ i ][ j ] = nil
                end
            end
        end
    end
end

----------------------------------
-- Skill Tooltip
----------------------------------
function TCS:SkillTooltipInitialize()
    -- Function prehooks
    ZO_PreHook(
        "ZO_AbilityTooltip_OnAddGameData",
        function()
            self.toolTipInitializeCounter = self.toolTipInitializeCounter + 1

            if( self.toolTipInitializeCounter == 2 ) then
                self:SkillTooltipUpdateTooltip( AbilityTooltip )
            end
        end
    )

    ZO_PreHook(
        "ZO_SkillTooltip_OnAddGameData",
        function()
            self.toolTipInitializeCounter = self.toolTipInitializeCounter + 1

            if( self.toolTipInitializeCounter == 2 ) then
                self:SkillTooltipUpdateTooltip( SkillTooltip )
            end
        end
    )

    -- Register handlers
    ZO_PreHookHandler(
        self.control, "OnUpdate",
        function() self:OnUpdate() end
    )
end

function TCS:SkillTooltipOnUpdate()
    -- execture when the mouseover has changed
    if( WINDOW_MANAGER:GetMouseOverControl() ~= self.lastMouseOver ) then
        self.lastMouseOver = WINDOW_MANAGER:GetMouseOverControl()
        self.toolTipInitializeCounter = 0
    end
end

function TCS:SkillTooltipUpdateTooltip( _tooltip )
    _tooltip:AddVerticalPadding( 3 )

    local control = WINDOW_MANAGER:GetMouseOverControl()

    -- Get the ability name
    local abilityNameIndex

    if( control.slotNum ~= nil ) then abilityNameIndex = zo_strformat( "<<z:1>>", GetSlotName( control.slotNum ) ) end
    if( control.ability ~= nil and control.ability.purchased ) then abilityNameIndex = zo_strformat( "<<z:1>>", control.ability.plainName ) end

    if( abilityNameIndex == nil or #abilityNameIndex == 0 ) then
        return
    end

    abilityNameIndex = self.abilityNameProxies[ abilityNameIndex ] or abilityNameIndex

    -- Dont do anything if there is nothing to display
    if( self.settings.abilityStatistics[ abilityNameIndex ] == nil or not self.settings.abilityStatisticOptions.enableTooltip or (
        self.settings.abilityStatistics[ abilityNameIndex ].normalAverage   == 0 and self.settings.abilityStatistics[ abilityNameIndex ].normalHighest      == 0 and
        self.settings.abilityStatistics[ abilityNameIndex ].normalLowest    == 0 and self.settings.abilityStatistics[ abilityNameIndex ].criticalAverage    == 0 and
        self.settings.abilityStatistics[ abilityNameIndex ].criticalHighest == 0 and self.settings.abilityStatistics[ abilityNameIndex ].criticalLowest     == 0
        )
    ) then
        return
    end

    -- Display the statistics
    local paddingNeeded = false

    if( tonumber( self.settings.abilityStatistics[ abilityNameIndex ].normalAverage ) ~= 0 and self.settings.abilityStatisticOptions.showNormalAverage ) then
        self:SkillTooltipAddLine( _tooltip, zo_strformat( GetString( SI_TCS_GENERIC_AVERAGE ), self.settings.abilityStatistics[ abilityNameIndex ].damageType ), zo_round( tonumber( self.settings.abilityStatistics[ abilityNameIndex ].normalAverage ) ) )

        paddingNeeded = true
    end

    if( tonumber( self.settings.abilityStatistics[ abilityNameIndex ].normalAverage ) ~= 0 and self.settings.abilityStatisticOptions.showNormalAverage and
        tonumber( self.settings.abilityStatistics[ abilityNameIndex ].normalHighest ) ~= 0 and self.settings.abilityStatisticOptions.showNormalMaximum
    ) then
        self:SkillTooltipAddLine(   _tooltip,   "     " .. zo_strformat( GetString( SI_TCS_GENERIC_MAXIMUM ), self.settings.abilityStatistics[ abilityNameIndex ].damageType ), tonumber( self.settings.abilityStatistics[ abilityNameIndex ].normalHighest ) )

        paddingNeeded = true
    end

    if( tonumber( self.settings.abilityStatistics[ abilityNameIndex ].normalAverage ) ~= 0 and self.settings.abilityStatisticOptions.showNormalAverage and
        tonumber( self.settings.abilityStatistics[ abilityNameIndex ].normalLowest ) ~= 0 and self.settings.abilityStatisticOptions.showNormalMinimum
    ) then
        self:SkillTooltipAddLine(   _tooltip,   "     " .. zo_strformat( GetString( SI_TCS_GENERIC_MINIMUM ), self.settings.abilityStatistics[ abilityNameIndex ].damageType ), tonumber( self.settings.abilityStatistics[ abilityNameIndex ].normalLowest ) )

        paddingNeeded = true
    end

    if( not self.settings.abilityStatisticOptions.showNormalAverage and
        tonumber( self.settings.abilityStatistics[ abilityNameIndex ].normalHighest ) ~= 0 and self.settings.abilityStatisticOptions.showNormalMaximum
    ) then
        self:SkillTooltipAddLine(   _tooltip,   zo_strformat( GetString( SI_TCS_GENERIC_MAXIMUM ), self.settings.abilityStatistics[ abilityNameIndex ].damageType ), tonumber( self.settings.abilityStatistics[ abilityNameIndex ].normalHighest ) )

        paddingNeeded = true
    end

    if( not self.settings.abilityStatisticOptions.showNormalAverage and
        tonumber( self.settings.abilityStatistics[ abilityNameIndex ].normalLowest ) ~= 0 and self.settings.abilityStatisticOptions.showNormalMinimum
    ) then
        self:SkillTooltipAddLine(   _tooltip,   zo_strformat( GetString( SI_TCS_GENERIC_MINIMUM ), self.settings.abilityStatistics[ abilityNameIndex ].damageType ), tonumber( self.settings.abilityStatistics[ abilityNameIndex ].normalLowest ) )

        paddingNeeded = true
    end

    if( paddingNeeded ) then
        _tooltip:AddVerticalPadding( 8 )
    end

    paddingNeeded = false
    if( tonumber( self.settings.abilityStatistics[ abilityNameIndex ].criticalAverage ) ~= 0 and self.settings.abilityStatisticOptions.showCriticalAverage ) then
        self:SkillTooltipAddLine(   _tooltip,   zo_strformat( GetString( SI_TCS_GENERIC_AVERAGE_CRITICAL ), self.settings.abilityStatistics[ abilityNameIndex ].damageType ), zo_round( tonumber( self.settings.abilityStatistics[ abilityNameIndex ].criticalAverage ) ) )
        paddingNeeded = true
    end

    if( tonumber( self.settings.abilityStatistics[ abilityNameIndex ].criticalAverage ) ~= 0 and self.settings.abilityStatisticOptions.showCriticalAverage and
        tonumber( self.settings.abilityStatistics[ abilityNameIndex ].criticalHighest ) ~= 0 and self.settings.abilityStatisticOptions.showCriticalMaximum
    ) then
        self:SkillTooltipAddLine(   _tooltip,   "     " .. zo_strformat( GetString( SI_TCS_GENERIC_MAXIMUM_CRITICAL ), self.settings.abilityStatistics[ abilityNameIndex ].damageType ), tonumber( self.settings.abilityStatistics[ abilityNameIndex ].criticalHighest ) )
        paddingNeeded = true
    end

    if( tonumber( self.settings.abilityStatistics[ abilityNameIndex ].criticalAverage ) ~= 0 and self.settings.abilityStatisticOptions.showCriticalAverage and
        tonumber( self.settings.abilityStatistics[ abilityNameIndex ].criticalLowest ) ~= 0 and self.settings.abilityStatisticOptions.showCriticalMinimum
    ) then
        self:SkillTooltipAddLine(   _tooltip,   "     " .. zo_strformat( GetString( SI_TCS_GENERIC_MINIMUM ), self.settings.abilityStatistics[ abilityNameIndex ].damageType ), tonumber( self.settings.abilityStatistics[ abilityNameIndex ].criticalLowest ) )
        paddingNeeded = true
    end

    if( not self.settings.abilityStatisticOptions.showCriticalAverage and
        tonumber( self.settings.abilityStatistics[ abilityNameIndex ].criticalHighest ) ~= 0 and self.settings.abilityStatisticOptions.showCriticalMaximum
    ) then
        self:SkillTooltipAddLine(   _tooltip,   zo_strformat( GetString( SI_TCS_GENERIC_MAXIMUM_CRITICAL ), self.settings.abilityStatistics[ abilityNameIndex ].damageType ), tonumber( self.settings.abilityStatistics[ abilityNameIndex ].criticalHighest ) )
        paddingNeeded = true
    end

    if( not self.settings.abilityStatisticOptions.showCriticalAverage and
        tonumber( self.settings.abilityStatistics[ abilityNameIndex ].criticalLowest ) ~= 0 and self.settings.abilityStatisticOptions.showCriticalMinimum
    ) then
        self:SkillTooltipAddLine(   _tooltip,   zo_strformat( GetString( SI_TCS_GENERIC_MINIMUM ), self.settings.abilityStatistics[ abilityNameIndex ].damageType ), tonumber( self.settings.abilityStatistics[ abilityNameIndex ].criticalLowest ) )
        paddingNeeded = true
    end

    if( paddingNeeded ) then
        _tooltip:AddVerticalPadding( 8 )
    end

    local normalAverage     = tonumber( self.settings.abilityStatistics[ abilityNameIndex ].normalAverage )
    local criticalAverage   = tonumber( self.settings.abilityStatistics[ abilityNameIndex ].criticalAverage )
    local totalAverage      = 0

    if( normalAverage == 0 ) then
        totalAverage = criticalAverage
    elseif( criticalAverage == 0 ) then
        totalAverage = normalAverage
    else
        totalAverage = ( normalAverage + criticalAverage ) / 2
    end

    if( self.settings.abilityStatisticOptions.showTotalAverage ) then
        self:SkillTooltipAddLine(   _tooltip,   zo_strformat( GetString( SI_TCS_GENERIC_TOTAL_AVERAGE ), self.settings.abilityStatistics[ abilityNameIndex ].damageType ), zo_round( totalAverage ) )
    end

    if( self.settings.abilityStatisticOptions.showCriticalPercent ) then
        self:SkillTooltipAddLine(   _tooltip,   GetString( SI_TCS_GENERIC_CRITICAL_CHANCE ), ( "%04.02f%%" ):format( zo_round( self.settings.abilityStatistics[ abilityNameIndex ].criticalPercent * 100 ) / 100 ) )
    end

    if( self.settings.abilityStatisticOptions.showNormalAverage or      self.settings.abilityStatisticOptions.showNormalMaximum or
        self.settings.abilityStatisticOptions.showNormalMinimum or      self.settings.abilityStatisticOptions.showCriticalAverage or
        self.settings.abilityStatisticOptions.showCriticalMaximum or    self.settings.abilityStatisticOptions.showCriticalMinimum or
        self.settings.abilityStatisticOptions.showTotalAverage or       self.settings.abilityStatisticOptions.showCriticalPercent
    ) then
        _tooltip:AddVerticalPadding( 10 )
    end
end

function TCS:SkillTooltipAddLine( _tooltip, _valueName, _value )
    local valueName = ZO_TOOLTIP_DEFAULT_COLOR:Colorize( _valueName )
    local value     = tostring( _value )

    _tooltip:AddLine( valueName, self.TOOLTIP_FONT, 1, 1, 1, nil, nil, TEXT_ALIGN_LEFT, true )

    _tooltip:AddVerticalPadding( -25 )
    _tooltip:AddLine( value, self.TOOLTIP_FONT, 1, 1, 1, nil, nil, TEXT_ALIGN_RIGHT, true )

    _tooltip:AddVerticalPadding( 1 )
end

----------------------------------
-- Combat Session
----------------------------------
function TCS:CombatSessionDisplayInitialize()
    -- DPS
    self.combatSessionDPSDisplay = WINDOW_MANAGER:CreateControl( "TCS_Combat_Session_DPS_Display", ZO_ActionBar1, CT_TEXTURE )
    self.combatSessionDPSDisplay:SetDimensions( 121, 20 )
    self.combatSessionDPSDisplay:SetMouseEnabled( true )
    self.combatSessionDPSDisplay:SetMovable( true )
    self.combatSessionDPSDisplay:SetAnchor(
        self.settings.combatSessionDPSDisplayPoint,
        _G[self.settings.combatSessionDPSDisplayRelativeTo],
        self.settings.combatSessionDPSDisplayRelativePoint,
        self.settings.combatSessionDPSDisplayOffsetX,
        self.settings.combatSessionDPSDisplayOffsetY
    )

    self.combatSessionDPSDisplay.dps = WINDOW_MANAGER:CreateControl( "TCS_Combat_Session_Display_DPS_Label", ZO_ActionBar1, CT_LABEL )
    self.combatSessionDPSDisplay.dps:SetAnchor( CENTER, TCS_Combat_Session_DPS_Display, CENTER, 0, -1 )
    self.combatSessionDPSDisplay.dps:SetText( GetString( SI_TCS_GENERIC_DPS ) .. ": 0" )

    self:SetCombatSessionDPSUpdateHidden()

    -- Total Damage
    self.combatSessionTotalDamageDisplay = WINDOW_MANAGER:CreateControl( "TCS_Combat_Session_Total_Damage_Display", ZO_ActionBar1, CT_TEXTURE )
    self.combatSessionTotalDamageDisplay:SetDimensions( 121, 20 )
    self.combatSessionTotalDamageDisplay:SetMouseEnabled( true )
    self.combatSessionTotalDamageDisplay:SetMovable( true )
    self.combatSessionTotalDamageDisplay:SetAnchor(
        self.settings.combatSessionTotalDamageDisplayPoint,
        _G[self.settings.combatSessionTotalDamageDisplayRelativeTo],
        self.settings.combatSessionTotalDamageDisplayRelativePoint,
        self.settings.combatSessionTotalDamageDisplayOffsetX,
        self.settings.combatSessionTotalDamageDisplayOffsetY
    )

    self.combatSessionTotalDamageDisplay.totalDamage = WINDOW_MANAGER:CreateControl( "TCS_Combat_Session_Display_Total_Damage_Label", ZO_ActionBar1, CT_LABEL )
    self.combatSessionTotalDamageDisplay.totalDamage:SetAnchor( CENTER, TCS_Combat_Session_Total_Damage_Display, CENTER, 0, -1 )
    self.combatSessionTotalDamageDisplay.totalDamage:SetText( GetString( SI_TCS_SETTINGS_DAMAGE_TOTAL_LABEL ) .. ": 0" )

    self:SetCombatSessionTotalDamageUpdateHidden()

    -- HPS
    self.combatSessionHPSDisplay = WINDOW_MANAGER:CreateControl( "TCS_Combat_Session_HPS_Display", ZO_ActionBar1, CT_TEXTURE )
    self.combatSessionHPSDisplay:SetDimensions( 121, 20 )
    self.combatSessionHPSDisplay:SetMouseEnabled( true )
    self.combatSessionHPSDisplay:SetMovable( true )
    self.combatSessionHPSDisplay:SetAnchor(
        self.settings.combatSessionHPSDisplayPoint,
        _G[self.settings.combatSessionHPSDisplayRelativeTo],
        self.settings.combatSessionHPSDisplayRelativePoint,
        self.settings.combatSessionHPSDisplayOffsetX,
        self.settings.combatSessionHPSDisplayOffsetY
    )

    self.combatSessionHPSDisplay.hps = WINDOW_MANAGER:CreateControl( "TCS_Combat_Session_Display_HPS_Label", ZO_ActionBar1, CT_LABEL )
    self.combatSessionHPSDisplay.hps:SetAnchor( CENTER, TCS_Combat_Session_HPS_Display, CENTER, 0, -1 )
    self.combatSessionHPSDisplay.hps:SetText( GetString( SI_TCS_GENERIC_HPS ) .. ": 0" )

    self:SetCombatSessionHPSUpdateHidden()

    -- Total Healing
    self.combatSessionTotalHealingDisplay = WINDOW_MANAGER:CreateControl( "TCS_Combat_Session_Total_Healing_Display", ZO_ActionBar1, CT_TEXTURE )
    self.combatSessionTotalHealingDisplay:SetDimensions( 121, 20 )
    self.combatSessionTotalHealingDisplay:SetMouseEnabled( true )
    self.combatSessionTotalHealingDisplay:SetMovable( true )
    self.combatSessionTotalHealingDisplay:SetAnchor(
        self.settings.combatSessionTotalHealingDisplayPoint,
        _G[self.settings.combatSessionTotalHealingDisplayRelativeTo],
        self.settings.combatSessionTotalHealingDisplayRelativePoint,
        self.settings.combatSessionTotalHealingDisplayOffsetX,
        self.settings.combatSessionTotalHealingDisplayOffsetY
    )

    self.combatSessionTotalHealingDisplay.totalHealing = WINDOW_MANAGER:CreateControl( "TCS_Combat_Session_Display_Total_Healing_Label", ZO_ActionBar1, CT_LABEL )
    self.combatSessionTotalHealingDisplay.totalHealing:SetAnchor( CENTER, TCS_Combat_Session_Total_Healing_Display, CENTER, 0, -1 )
    self.combatSessionTotalHealingDisplay.totalHealing:SetText( GetString( SI_TCS_SETTINGS_HEAL_TOTAL_SHOW_LABEL ) .. ": 0" )

    self:SetCombatSessionTotalHealingUpdateHidden()

    self:SetCombatSessionDisplayBGType( self.settings.combatSessionDisplayBGType )
    self:SetCombatSessionDisplayBGTransparency( self.settings.combatSessionDisplayBGTransparency )
    self:SetCombatSessionDisplayTextTransparency( self.settings.combatSessionDisplayTextTransparency )

    self:UpdateCombatSessionDisplays()

    -- Show action bar when changing settings
    ZO_PreHookHandler(
        TCS_SettingsMenu_BG_Type_Combat_Session_Display_Dropdown, "OnMouseEnter",
        function() ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_BG_Type_Combat_Session_Display_Dropdown, "OnMouseExit",
        function() ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_BG_Type_Combat_Session_Display_Dropdown.dropdown:GetControl(), "OnMouseEnter",
        function() ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_BG_Type_Combat_Session_Display_Dropdown.dropdown:GetControl(), "OnMouseExit",
        function() ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_Scale_Combat_Session_Display_Slider, "OnMouseEnter",
        function() ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_Scale_Combat_Session_Display_Slider, "OnMouseExit",
        function() ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_Scale_Combat_Session_Display_Slider.slider, "OnMouseEnter",
        function() ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_Scale_Combat_Session_Display_Slider.slider, "OnMouseExit",
        function() ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_BG_Transparency_Combat_Session_Display_Slider, "OnMouseEnter",
        function() ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_BG_Transparency_Combat_Session_Display_Slider, "OnMouseExit",
        function() ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_BG_Transparency_Combat_Session_Display_Slider.slider, "OnMouseEnter",
        function() ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_BG_Transparency_Combat_Session_Display_Slider.slider, "OnMouseExit",
        function() ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_Text_Transparency_Combat_Session_Display_Slider, "OnMouseEnter",
        function() ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_Text_Transparency_Combat_Session_Display_Slider, "OnMouseExit",
        function() ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_Text_Transparency_Combat_Session_Display_Slider.slider, "OnMouseEnter",
        function() ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_Text_Transparency_Combat_Session_Display_Slider.slider, "OnMouseExit",
        function() ZO_ActionBar1:SetHidden( true ) end
    )
end

function TCS:SetCombatSessionDisplayFontScale( _scale )
    local scale = _scale

    if( scale < 1.0 ) then scale = 1 end
    if( scale > 6.0 ) then scale = 6 end

    if( scale == 1 ) then
        self.combatSessionDPSDisplay.dps:SetFont( "ZoFontGameSmall" )
        self.combatSessionTotalDamageDisplay.totalDamage:SetFont( "ZoFontGameSmall" )
        self.combatSessionHPSDisplay.hps:SetFont( "ZoFontGameSmall" )
        self.combatSessionTotalHealingDisplay.totalHealing:SetFont( "ZoFontGameSmall" )
    elseif( scale == 2 ) then
        self.combatSessionDPSDisplay.dps:SetFont( "ZoFontWinT2" )
        self.combatSessionTotalDamageDisplay.totalDamage:SetFont( "ZoFontWinT2" )
        self.combatSessionHPSDisplay.hps:SetFont( "ZoFontWinT2" )
        self.combatSessionTotalHealingDisplay.totalHealing:SetFont( "ZoFontWinT2" )
    elseif( scale == 3 ) then
        self.combatSessionDPSDisplay.dps:SetFont( "ZoFontWinT1" )
        self.combatSessionTotalDamageDisplay.totalDamage:SetFont( "ZoFontWinT1" )
        self.combatSessionHPSDisplay.hps:SetFont( "ZoFontWinT1" )
        self.combatSessionTotalHealingDisplay.totalHealing:SetFont( "ZoFontWinT1" )
    elseif( scale == 4 ) then
        self.combatSessionDPSDisplay.dps:SetFont( "ZoFontWinH3" )
        self.combatSessionTotalDamageDisplay.totalDamage:SetFont( "ZoFontWinH3" )
        self.combatSessionHPSDisplay.hps:SetFont( "ZoFontWinH3" )
        self.combatSessionTotalHealingDisplay.totalHealing:SetFont( "ZoFontWinH3" )
    elseif( scale == 5 ) then
        self.combatSessionDPSDisplay.dps:SetFont( "ZoFontWinH2" )
        self.combatSessionTotalDamageDisplay.totalDamage:SetFont( "ZoFontWinH2" )
        self.combatSessionHPSDisplay.hps:SetFont( "ZoFontWinH2" )
        self.combatSessionTotalHealingDisplay.totalHealing:SetFont( "ZoFontWinH2" )
    elseif( scale == 6 ) then
        self.combatSessionDPSDisplay.dps:SetFont( "ZoFontWinH1" )
        self.combatSessionTotalDamageDisplay.totalDamage:SetFont( "ZoFontWinH1" )
        self.combatSessionHPSDisplay.hps:SetFont( "ZoFontWinH1" )
        self.combatSessionTotalHealingDisplay.totalHealing:SetFont( "ZoFontWinH1" )
    end
end

function TCS:SetCombatSessionDisplayBGScale( _scale )
    -- First set the width then the scale, very important!
    local scale = _scale

    if( scale < 1.0 ) then scale = 1 end
    if( scale > 6.0 ) then scale = 6 end

    if( scale == 1 ) then
        self.combatSessionDPSDisplay:SetScale( 1 )
        self.combatSessionTotalDamageDisplay:SetScale( 1 )
        self.combatSessionHPSDisplay:SetScale( 1 )
        self.combatSessionTotalHealingDisplay:SetScale( 1 )
    elseif( scale == 2 ) then
        self.combatSessionDPSDisplay:SetScale( 1.1 )
        self.combatSessionTotalDamageDisplay:SetScale( 1.1 )
        self.combatSessionHPSDisplay:SetScale( 1.1 )
        self.combatSessionTotalHealingDisplay:SetScale( 1.1 )
    elseif( scale == 3 ) then
        self.combatSessionDPSDisplay:SetScale( 1.25 )
        self.combatSessionTotalDamageDisplay:SetScale( 1.25 )
        self.combatSessionHPSDisplay:SetScale( 1.25 )
        self.combatSessionTotalHealingDisplay:SetScale( 1.25 )
    elseif( scale == 4 ) then
        self.combatSessionDPSDisplay:SetScale( 1.4 )
        self.combatSessionTotalDamageDisplay:SetScale( 1.4 )
        self.combatSessionHPSDisplay:SetScale( 1.4 )
        self.combatSessionTotalHealingDisplay:SetScale( 1.4 )
    elseif( scale == 5 ) then
        self.combatSessionDPSDisplay:SetScale( 1.65 )
        self.combatSessionTotalDamageDisplay:SetScale( 1.65 )
        self.combatSessionHPSDisplay:SetScale( 1.65 )
        self.combatSessionTotalHealingDisplay:SetScale( 1.65 )
    elseif( scale == 6 ) then
        self.combatSessionDPSDisplay:SetScale( 2.1 )
        self.combatSessionTotalDamageDisplay:SetScale( 2.1 )
        self.combatSessionHPSDisplay:SetScale( 2.1 )
        self.combatSessionTotalHealingDisplay:SetScale( 2.1 )
    end
end

function TCS:SetCombatSessionDisplayBGType( _bgType )
    local bgType = _bgType

    if( bgType == "" ) then
        self.settings.combatSessionDisplayBGType = GetString( SI_TCS_SETTINGS_BACKGROUND_TYPE_A )
        bgType = self.settings.combatSessionDisplayBGType
    end

    if( bgType == GetString( SI_TCS_SETTINGS_BACKGROUND_TYPE_A ) ) then
        self.combatSessionDPSDisplay:SetTexture( "TsCombatStats/Textures/display_bg_type_a.dds" )
        self.combatSessionTotalDamageDisplay:SetTexture( "TsCombatStats/Textures/display_bg_type_a.dds" )
        self.combatSessionHPSDisplay:SetTexture( "TsCombatStats/Textures/display_bg_type_a.dds" )
        self.combatSessionTotalHealingDisplay:SetTexture( "TsCombatStats/Textures/display_bg_type_a.dds" )
    elseif( bgType == GetString( SI_TCS_SETTINGS_BACKGROUND_TYPE_B ) ) then
        self.combatSessionDPSDisplay:SetTexture( "TsCombatStats/Textures/display_bg_type_b.dds" )
        self.combatSessionTotalDamageDisplay:SetTexture( "TsCombatStats/Textures/display_bg_type_b.dds" )
        self.combatSessionHPSDisplay:SetTexture( "TsCombatStats/Textures/display_bg_type_b.dds" )
        self.combatSessionTotalHealingDisplay:SetTexture( "TsCombatStats/Textures/display_bg_type_b.dds" )
    elseif( bgType == GetString( SI_TCS_SETTINGS_BACKGROUND_TYPE_NONE ) ) then
        self.combatSessionDPSDisplay:SetTexture( "TsCombatStats/Textures/display_bg_type_none.dds" )
        self.combatSessionTotalDamageDisplay:SetTexture( "TsCombatStats/Textures/display_bg_type_none.dds" )
        self.combatSessionHPSDisplay:SetTexture( "TsCombatStats/Textures/display_bg_type_none.dds" )
        self.combatSessionTotalHealingDisplay:SetTexture( "TsCombatStats/Textures/display_bg_type_none.dds" )
    end
end

function TCS:SetCombatSessionDisplayBGTransparency( _transparency )
    local transparency = _transparency / 100

    self.combatSessionDPSDisplay:SetAlpha( transparency )
    self.combatSessionTotalDamageDisplay:SetAlpha( transparency )
    self.combatSessionHPSDisplay:SetAlpha( transparency )
    self.combatSessionTotalHealingDisplay:SetAlpha( transparency )
end

function TCS:SetCombatSessionDisplayTextTransparency( _transparency )
    local transparency = _transparency / 100

    self.combatSessionDPSDisplay.dps:SetAlpha( transparency )
    self.combatSessionTotalDamageDisplay.totalDamage:SetAlpha( transparency )
    self.combatSessionHPSDisplay.hps:SetAlpha( transparency )
    self.combatSessionTotalHealingDisplay.totalHealing:SetAlpha( transparency )
end

function TCS:HideCombatSessionDisplays( _hide )
    self:SetCombatSessionDPSUpdateHidden( _hide )
    self:SetCombatSessionTotalDamageUpdateHidden( _hide )
    self:SetCombatSessionHPSUpdateHidden( _hide )
    self:SetCombatSessionTotalHealingUpdateHidden( _hide )
end

function TCS:SetCombatSessionDPSUpdateHidden( _hide )
    if( self.combatSessionDPSDisplay == nil ) then return end
    local isHidden = false

    if( _hide == true ) then
        isHidden = _hide
    else
        isHidden = not self.settings.showDPS
    end

    self.combatSessionDPSDisplay:SetHidden( isHidden )
    self.combatSessionDPSDisplay.dps:SetHidden( isHidden )
end

function TCS:SetCombatSessionTotalDamageUpdateHidden( _hide )
    if( self.combatSessionTotalDamageDisplay == nil ) then return end
    local isHidden = false

    if( _hide == true ) then
        isHidden = _hide
    else
        isHidden = not self.settings.totalDamage
    end

    self.combatSessionTotalDamageDisplay:SetHidden( isHidden )
    self.combatSessionTotalDamageDisplay.totalDamage:SetHidden( isHidden )
end

function TCS:SetCombatSessionHPSUpdateHidden( _hide )
    if( self.combatSessionHPSDisplay == nil ) then return end
    local isHidden = false

    if( _hide == true ) then
        isHidden = true
    else
        isHidden = not self.settings.showHPS
    end

    self.combatSessionHPSDisplay:SetHidden( isHidden )
    self.combatSessionHPSDisplay.hps:SetHidden( isHidden )
end

function TCS:SetCombatSessionTotalHealingUpdateHidden( _hide )
    if( self.combatSessionTotalHealingDisplay == nil ) then return end
    local isHidden = false

    if( _hide == true ) then
        isHidden = _hide
    else
        isHidden = not self.settings.totalHealing
    end

    self.combatSessionTotalHealingDisplay:SetHidden( isHidden )
    self.combatSessionTotalHealingDisplay.totalHealing:SetHidden( isHidden )
end

function TCS:UpdateCombatSessionDisplays()
    self:SetCombatSessionDisplayFontScale( self.settings.combatSessionDisplayScale )

    -- First set the width then the scale, very important!
    local scale = self.settings.combatSessionDisplayScale

    if( self.combatSessionStartTime == 0 ) then self.combatSessionStartTime = GetGameTimeMilliseconds() end

    if( self.settings.showDPS ) then
        local currentDPS    = 0
        local timeElapsed   = GetGameTimeMilliseconds() - self.combatSessionStartTime

        if( timeElapsed ~= 0 ) then
            currentDPS = self.combatSessionDamageDealt / ( timeElapsed / 1000 )
        end

        local text = GetString( SI_TCS_GENERIC_DPS ) .. ": " .. tostring( zo_round( currentDPS ) )
        local textWidth = self.combatSessionDPSDisplay.dps:GetStringWidth( text )

        if( scale == 1 ) then
            textWidth = textWidth + ( ( zo_pow( textWidth, 1.022 ) - textWidth ) + 12 )
        elseif( scale == 2 ) then
            textWidth = textWidth + 12
        elseif( scale == 3 ) then
            textWidth = textWidth + ( ( zo_pow( textWidth, 0.982 ) - textWidth ) + 5 )
        elseif( scale == 4 ) then
            textWidth = textWidth + ( ( zo_pow( textWidth, 0.966 ) - textWidth ) + 5 )
        elseif( scale == 5 ) then
            textWidth = textWidth + ( zo_pow( textWidth, 0.94 ) - textWidth )
        elseif( scale == 6 ) then
            textWidth = textWidth + ( zo_pow( textWidth, 0.905 ) - textWidth )
        end

        self.combatSessionDPSDisplay.dps:SetText( text )
        self.combatSessionDPSDisplay:SetDimensions( textWidth, 20 )
    end

    if( self.settings.showHPS ) then
        local currentHPS    = 0
        local timeElapsed   = GetGameTimeMilliseconds() - self.combatSessionStartTime

        if( timeElapsed ~= 0 ) then
            currentHPS = self.combatSessionHealingDone / ( timeElapsed / 1000 )
        end

        local text = GetString( SI_TCS_GENERIC_HPS ) .. ": " .. tostring( zo_round( currentHPS ) )
        local textWidth = self.combatSessionHPSDisplay.hps:GetStringWidth( text )

        if( scale == 1 ) then
            textWidth = textWidth + ( ( zo_pow( textWidth, 1.022 ) - textWidth ) + 12 )
        elseif( scale == 2 ) then
            textWidth = textWidth + 12
        elseif( scale == 3 ) then
            textWidth = textWidth + ( ( zo_pow( textWidth, 0.982 ) - textWidth ) + 5 )
        elseif( scale == 4 ) then
            textWidth = textWidth + ( ( zo_pow( textWidth, 0.966 ) - textWidth ) + 5 )
        elseif( scale == 5 ) then
            textWidth = textWidth + ( zo_pow( textWidth, 0.94 ) - textWidth )
        elseif( scale == 6 ) then
            textWidth = textWidth + ( zo_pow( textWidth, 0.905 ) - textWidth )
        end

        self.combatSessionHPSDisplay.hps:SetText( text )
        self.combatSessionHPSDisplay:SetDimensions( textWidth, 20 )
    end

    if( self.settings.totalDamage ) then
        local text = GetString( SI_TCS_SETTINGS_DAMAGE_TOTAL_LABEL ) .. ": " .. tostring( zo_round( self.combatSessionDamageDealt ) )
        local textWidth = self.combatSessionTotalDamageDisplay.totalDamage:GetStringWidth( text )

        if( scale == 1 ) then
            textWidth = textWidth + ( ( zo_pow( textWidth, 1.022 ) - textWidth ) + 12 )
        elseif( scale == 2 ) then
            textWidth = textWidth + 12
        elseif( scale == 3 ) then
            textWidth = textWidth + ( ( zo_pow( textWidth, 0.982 ) - textWidth ) + 5 )
        elseif( scale == 4 ) then
            textWidth = textWidth + ( ( zo_pow( textWidth, 0.966 ) - textWidth ) + 5 )
        elseif( scale == 5 ) then
            textWidth = textWidth + ( zo_pow( textWidth, 0.94 ) - textWidth )
        elseif( scale == 6 ) then
            textWidth = textWidth + ( zo_pow( textWidth, 0.905 ) - textWidth )
        end

        self.combatSessionTotalDamageDisplay.totalDamage:SetText( text )
        self.combatSessionTotalDamageDisplay:SetDimensions( textWidth, 20 )
    end

    if( self.settings.totalHealing ) then
        local text = GetString( SI_TCS_SETTINGS_HEAL_TOTAL_SHOW_LABEL ) .. ": " .. tostring( zo_round( self.combatSessionHealingDone ) )
        local textWidth = self.combatSessionTotalHealingDisplay.totalHealing:GetStringWidth( text )

        if( scale == 1 ) then
            textWidth = textWidth + ( ( zo_pow( textWidth, 1.022 ) - textWidth ) + 12 )
        elseif( scale == 2 ) then
            textWidth = textWidth + 12
        elseif( scale == 3 ) then
            textWidth = textWidth + ( ( zo_pow( textWidth, 0.982 ) - textWidth ) + 5 )
        elseif( scale == 4 ) then
            textWidth = textWidth + ( ( zo_pow( textWidth, 0.966 ) - textWidth ) + 5 )
        elseif( scale == 5 ) then
            textWidth = textWidth + ( zo_pow( textWidth, 0.94 ) - textWidth )
        elseif( scale == 6 ) then
            textWidth = textWidth + ( zo_pow( textWidth, 0.905 ) - textWidth )
        end

        self.combatSessionTotalHealingDisplay.totalHealing:SetText( text )
        self.combatSessionTotalHealingDisplay:SetDimensions( textWidth, 20 )
    end

    self:SetCombatSessionDisplayBGScale( self.settings.combatSessionDisplayScale )
end

----------------------------------
-- Ability Damage Percentages
----------------------------------
function TCS:InitializeAbilityDamagePercentages()
    self.control:RegisterForEvent(
        EVENT_ACTION_SLOT_UPDATED,
        function( _eventCode, _slotNum )
            if( not self.isInCombat and _slotNum > 2 and _slotNum < 9 ) then
                self:ResetAbilityDamagePercentage()
                self:UpdateSlotNameCache()
            end
        end
    )

    self:UpdateSlotNameCache()

    for i = 1, 6 do
        local textureName = "TCS_Combat_Session_Ability_Damage_Percentage" .. i
        local labelName = "TCS_Combat_Session_Ability_Damage_PercentageText" .. i
        local parent = WINDOW_MANAGER:GetControlByName( "ActionButton" .. i+2 .. "Button" )

        WINDOW_MANAGER:CreateControl( textureName, parent , CT_TEXTURE )
        _G[ textureName ]:SetDrawLayer( 1 )

        WINDOW_MANAGER:CreateControl( labelName , parent, CT_LABEL )
        _G[ labelName ]:SetDrawLayer( 2 )
    end

    self:UpdateAbilityDamagePercentagesEnabled()
    self:UpdateAbilityDamagePercentagesWidth()
    self:UpdateAbilityDamagePercentagesColor()
    self:UpdateAbilityDamagePercentagesAlignment()

    self:UpdateAbilityDamagePercentagesTextEnabled()
    self:UpdateAbilityDamagePercentagesTextSize()
    self:UpdateAbilityDamagePercentagesTextAlignment()

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Enable_Bar_Indicators_Checkbox, "OnMouseEnter",
        function() self:SetAbilityDamagePercentagesPreview( true ) ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Enable_Bar_Indicators_Checkbox, "OnMouseExit",
        function() self:SetAbilityDamagePercentagesPreview( false ) ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Enable_Bar_Indicators_Checkbox.checkbox, "OnMouseEnter",
        function() self:SetAbilityDamagePercentagesPreview( true ) ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Enable_Bar_Indicators_Checkbox.checkbox, "OnMouseExit",
        function() self:SetAbilityDamagePercentagesPreview( false ) ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Bar_Indicators_Width_Slider, "OnMouseEnter",
        function() self:SetAbilityDamagePercentagesPreview( true ) ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Bar_Indicators_Width_Slider, "OnMouseExit",
        function() self:SetAbilityDamagePercentagesPreview( false ) ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Bar_Indicators_Width_Slider.slider, "OnMouseEnter",
        function() self:SetAbilityDamagePercentagesPreview( true ) ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Bar_Indicators_Width_Slider.slider, "OnMouseExit",
        function() self:SetAbilityDamagePercentagesPreview( false ) ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Bar_Indicators_Alignment_Dropdown, "OnMouseEnter",
        function() self:SetAbilityDamagePercentagesPreview( true ) ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Bar_Indicators_Alignment_Dropdown, "OnMouseExit",
        function() self:SetAbilityDamagePercentagesPreview( false ) ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Bar_Indicators_Alignment_Dropdown.dropdown:GetControl(), "OnMouseEnter",
        function() self:SetAbilityDamagePercentagesPreview( true ) ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Bar_Indicators_Alignment_Dropdown.dropdown:GetControl(), "OnMouseExit",
        function() self:SetAbilityDamagePercentagesPreview( false ) ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Bar_Indicators_Color_Colorpicker, "OnMouseEnter",
        function() self:SetAbilityDamagePercentagesPreview( true ) ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Bar_Indicators_Color_Colorpicker, "OnMouseExit",
        function() self:SetAbilityDamagePercentagesPreview( false ) ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Enable_Percentage_Indicators_Checkbox, "OnMouseEnter",
        function() self:SetAbilityDamagePercentagesPreview( true ) ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Enable_Percentage_Indicators_Checkbox, "OnMouseExit",
        function() self:SetAbilityDamagePercentagesPreview( false ) ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Enable_Percentage_Indicators_Checkbox.checkbox, "OnMouseEnter",
        function() self:SetAbilityDamagePercentagesPreview( true ) ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Enable_Percentage_Indicators_Checkbox.checkbox, "OnMouseExit",
        function() self:SetAbilityDamagePercentagesPreview( false ) ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Percentage_Indicators_Size_Slider, "OnMouseEnter",
        function() self:SetAbilityDamagePercentagesPreview( true ) ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Percentage_Indicators_Size_Slider, "OnMouseExit",
        function() self:SetAbilityDamagePercentagesPreview( false ) ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Percentage_Indicators_Size_Slider.slider, "OnMouseEnter",
        function() self:SetAbilityDamagePercentagesPreview( true ) ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Percentage_Indicators_Size_Slider.slider, "OnMouseExit",
        function() self:SetAbilityDamagePercentagesPreview( false ) ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Percentage_Indicators_Alignment_Dropdown, "OnMouseEnter",
        function() self:SetAbilityDamagePercentagesPreview( true ) ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Percentage_Indicators_Alignment_Dropdown, "OnMouseExit",
        function() self:SetAbilityDamagePercentagesPreview( false ) ZO_ActionBar1:SetHidden( true ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Percentage_Indicators_Alignment_Dropdown.dropdown:GetControl(), "OnMouseEnter",
        function() self:SetAbilityDamagePercentagesPreview( true ) ZO_ActionBar1:SetHidden( false ) end
    )

    ZO_PreHookHandler(
        TCS_SettingsMenu_AbilityDamagePercentages_Percentage_Indicators_Alignment_Dropdown.dropdown:GetControl(), "OnMouseExit",
        function() self:SetAbilityDamagePercentagesPreview( false ) ZO_ActionBar1:SetHidden( true ) end
    )
end

--_debugSlotNameCache = {}
function TCS:UpdateSlotNameCache()

    for i = 3, 8 do
        local abilityName = zo_strformat( "<<z:1>>", GetSlotName( i ) )
        abilityName = self.abilityNameProxies[ abilityName ] or abilityName

        self.slotNameCache[ i ] = abilityName
    end

    --_debugSlotNameCache = self.slotNameCache
end

function TCS:UpdateAbilityDamagePercentagesEnabled()
    local enabled = self.settings.combatSessionAbilityDamageEnableBarIndicators

    for i = 1, 6 do
        WINDOW_MANAGER:GetControlByName( "TCS_Combat_Session_Ability_Damage_Percentage" .. i ):SetHidden( not enabled )
    end
end

function TCS:UpdateAbilityDamagePercentagesWidth( _preview )
    local slotWidth = ( ActionButton3Button:GetWidth() - 4 ) * ( self.settings.combatSessionAbilityDamageBarWidth / 100 )
    local slotHeight = 0

    if( _preview ) then slotHeight = ActionButton3Button:GetHeight() - 4 end

    for i = 1, 6 do
        WINDOW_MANAGER:GetControlByName( "TCS_Combat_Session_Ability_Damage_Percentage" .. i ):SetDimensions( slotWidth, slotHeight )
    end
end

function TCS:UpdateAbilityDamagePercentagesColor()
    local r = self.settings.combatSessionAbilityDamageBarR
    local g = self.settings.combatSessionAbilityDamageBarG
    local b = self.settings.combatSessionAbilityDamageBarB
    local a = 0.75

    for i = 1, 6 do
        WINDOW_MANAGER:GetControlByName( "TCS_Combat_Session_Ability_Damage_Percentage" .. i ):SetColor( r, g, b, a )
    end
end

function TCS:UpdateAbilityDamagePercentagesAlignment()
    local alignment = self.settings.combatSessionAbilityDamageBarAlignment
    local offsetX = 2
    local offsetY = -2

    for i = 1, 6 do
        local control = WINDOW_MANAGER:GetControlByName( "TCS_Combat_Session_Ability_Damage_Percentage" .. i )
        local parent  = WINDOW_MANAGER:GetControlByName( "ActionButton" .. i+2 )
        control:ClearAnchors()

        if( alignment == GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_BAR_ALIGNMENT_LEFT ) ) then
            control:SetAnchor( BOTTOMLEFT, parent, BOTTOMLEFT, offsetX, offsetY )
        elseif( alignment == GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_BAR_ALIGNMENT_RIGHT ) ) then
            control:SetAnchor( BOTTOMRIGHT, parent, BOTTOMRIGHT, offsetX * -1, offsetY )
        end
    end
end

function TCS:UpdateAbilityDamagePercentagesTextEnabled()
    local enabled = self.settings.combatSessionAbilityDamageEnablePercentage

    if( ActionButton3Button.actionId ~= 0 ) then
        TCS_Combat_Session_Ability_Damage_PercentageText1:SetHidden( not enabled )
    else
        TCS_Combat_Session_Ability_Damage_PercentageText1:SetHidden( true )
    end

    if( ActionButton4Button.actionId ~= 0 ) then
        TCS_Combat_Session_Ability_Damage_PercentageText2:SetHidden( not enabled )
    else
        TCS_Combat_Session_Ability_Damage_PercentageText2:SetHidden( true )
    end

    if( ActionButton5Button.actionId ~= 0 ) then
        TCS_Combat_Session_Ability_Damage_PercentageText3:SetHidden( not enabled )
    else
        TCS_Combat_Session_Ability_Damage_PercentageText3:SetHidden( true )
    end

    if( ActionButton6Button.actionId ~= 0 ) then
        TCS_Combat_Session_Ability_Damage_PercentageText4:SetHidden( not enabled )
    else
        TCS_Combat_Session_Ability_Damage_PercentageText4:SetHidden( true )
    end

    if( ActionButton7Button.actionId ~= 0 ) then
        TCS_Combat_Session_Ability_Damage_PercentageText5:SetHidden( not enabled )
    else
        TCS_Combat_Session_Ability_Damage_PercentageText5:SetHidden( true )
    end

    if( ActionButton8Button.actionId ~= 0 ) then
        TCS_Combat_Session_Ability_Damage_PercentageText6:SetHidden( not enabled )
    else
        TCS_Combat_Session_Ability_Damage_PercentageText6:SetHidden( true )
    end
end

function TCS:UpdateAbilityDamagePercentagesTextSize()
    local size = self.settings.combatSessionAbilityDamagePercentageSize

    if( size < 1 ) then size = 1 end
    if( size > 4 ) then size = 4 end

    for i = 1, 6 do
        local control = WINDOW_MANAGER:GetControlByName( "TCS_Combat_Session_Ability_Damage_PercentageText" .. i )
        if( size == 1 ) then
            control:SetFont( "ZoFontGameSmall" )
        elseif( size == 2 ) then
            control:SetFont( "ZoFontWinT2" )
        elseif( size == 3 ) then
            control:SetFont( "ZoFontWinT1" )
        elseif( size == 4 ) then
            control:SetFont( "ZoFontWinH3" )
        end
    end
end

function TCS:UpdateAbilityDamagePercentagesTextAlignment()
    local alignment = self.settings.combatSessionAbilityDamagePercentageAlignment

    for i = 1, 6 do
        local control = WINDOW_MANAGER:GetControlByName( "TCS_Combat_Session_Ability_Damage_PercentageText" .. i )
        local parent = WINDOW_MANAGER:GetControlByName( "ActionButton" .. i+2 )
        local offsetX, offsetY, point = 0, 0

        if( alignment == GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_CENTER ) ) then
            point = CENTER
        elseif( alignment == GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_LEFT ) ) then
            offsetX = 4
            point = LEFT
        elseif( alignment == GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_RIGHT ) ) then
            offsetX = -4
            point = RIGHT
        elseif( alignment == GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_BOTTOMLEFT ) ) then
            offsetX = 4
            point = BOTTOMLEFT
        elseif( alignment == GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_BOTTOMRIGHT ) ) then
            offsetX = -4
            point = BOTTOMRIGHT
        elseif( alignment == GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_TOPLEFT ) ) then
            offsetX = 4
            point = TOPLEFT
        elseif( alignment == GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_TOPRIGHT ) ) then
            offsetX = -4
            point = TOPRIGHT
        elseif( alignment == GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_BOTTOM ) ) then
            point = BOTTOM
        elseif( alignment == GetString( SI_TCS_SETTINGS_ABILITY_DAMAGE_PERCENTAGES_TEXT_ALIGNMENT_TOP ) ) then
            point = TOP
        end

        control:ClearAnchors()
        control:SetAnchor( point, parent, point, offsetX, offsetY )

    end
end

function TCS:UpdateAbilityDamagePercentage( _abilityName, _damage )

    for i = 3, 8 do
        if( self.combatSessionAbilityDamage[ self.slotNameCache[ i ] ] == nil ) then
            self.combatSessionAbilityDamage[ self.slotNameCache[ i ] ] = 0
        end
    end

    local abilityName = zo_strformat( "<<z:1>>", _abilityName )
    abilityName = self.abilityNameProxies[ abilityName ] or abilityName

    if( self.combatSessionAbilityDamage[ abilityName ] == nil ) then return end

    self.combatSessionAbilityDamage[ abilityName ] = self.combatSessionAbilityDamage[ abilityName ] + _damage

    local slotHeight = ActionButton3Button:GetHeight() - 4
    local totalDamage = self.combatSessionDamageDealt
    local heighestDamage = 0

    for i = 3, 8 do
        local abilityDamage = self.combatSessionAbilityDamage[ self.slotNameCache[ i ] ]
        if( abilityDamage ~= nil and abilityDamage > 0 and abilityDamage > heighestDamage ) then
            heighestDamage = abilityDamage
        end
    end


    for i = 1, 6 do
        local texture = WINDOW_MANAGER:GetControlByName( "TCS_Combat_Session_Ability_Damage_Percentage" .. i )
        local label = WINDOW_MANAGER:GetControlByName( "TCS_Combat_Session_Ability_Damage_PercentageText" .. i )
        local damage = self.combatSessionAbilityDamage[ self.slotNameCache[ i+2 ] ]
        local slotPercentage = zo_round( ( damage / totalDamage ) * 100 )

        if( slotPercentage > 0 ) then
            texture:SetHeight( ( damage / heighestDamage ) * slotHeight )
            label:SetText( slotPercentage .. "%" )
        else
            texture:SetHeight( 0 )
            label:SetText( "" )
        end
    end

end

function TCS:ResetAbilityDamagePercentage()
    self.combatSessionAbilityDamage = {}

    for i = 1, 6 do
        WINDOW_MANAGER:GetControlByName( "TCS_Combat_Session_Ability_Damage_Percentage" .. i ):SetHeight( 0 )
        WINDOW_MANAGER:GetControlByName( "TCS_Combat_Session_Ability_Damage_PercentageText" .. i ):SetText( "" )
    end
end

function TCS:SetAbilityDamagePercentagesPreview( _preview )
    if( _preview ) then
        for i = 1, 6 do
            local texture = WINDOW_MANAGER:GetControlByName( "TCS_Combat_Session_Ability_Damage_Percentage" .. i )
            local label = WINDOW_MANAGER:GetControlByName( "TCS_Combat_Session_Ability_Damage_PercentageText" .. i )

            texture:SetHidden( false )
            texture:SetHeight( ActionButton3Button:GetHeight() - 4 )
            label:SetHidden( false )
            label:SetText( "100%" )
        end
    else
        self:UpdateAbilityDamagePercentagesEnabled()
        self:UpdateAbilityDamagePercentagesWidth()
        self:UpdateAbilityDamagePercentagesColor()
        self:UpdateAbilityDamagePercentagesAlignment()
        self:UpdateAbilityDamagePercentagesTextEnabled()
        self:UpdateAbilityDamagePercentagesTextSize()
        self:UpdateAbilityDamagePercentagesTextAlignment()
        self:ResetAbilityDamagePercentage()
    end
end